/*
 *      common.c - Common functions and functionalities used by usbsync
 * 		Part of the usbsync utility
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#else
#  define RETSIGTYPE void
#endif

#include "common.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <assert.h>
#include <getopt.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <libgen.h>
#include <sys/stat.h>

#ifdef HAVE_TERMIOS_H
#  include <termios.h>
#endif

#ifdef HAVE_TERMCAP
#  include <curses.h>
#  include <term.h>
#  include "status/status.h"
#endif

#include "error/usbs_errors.h"
#include "debug/debug.h"
#include "path.h"

/**
 * Does pretty much the opposite of strncmp in that it compares
 * the last n bytes of str1 and str2. Returns 0 if they match, -1
 * if they do not match. (In the last point this function differs
 * from strncmp in that way that no number of differing chars are
 * returned.)
 * NOTE: Both strings ought to be longer than one chars at least!
 */
int strrncmp(const char *str1, const char *str2, size_t n) {
	int i = 0;
	char *s1 = (char*)str1+strlen(str1)-1, *s2 = (char*)str2+strlen(str2)-1;
	/* Valid pointer check */
	if( str1 == NULL || str2 == NULL ) return -1;
	/* Dereferenced pointer check */
	if( !*str1 || !*str2  ) return -1;
	/* Check if last n chars match, char by char, return -1 if not */
	while(i++ < n) {
		if(s1 < str1 || s2 < str2) 
			return -1;
		if(*s1-- != *s2--) 
			return -1;
	}
	return 0;
}


/** 
 * This function supplies "stored strings".
 * They can be used by any function in order to have a reference to a dynamically
 * allocated string in order to escape the limiting dynamic stack.
 * A function may obtain such a reference by calling new_str( str ). Then str is 
 * copied into the newly allocated heap memory.
 * To clean up heap memory, new_str(NULL) should be called at frequent intervals.
 * Will return NULL if an error occured or if string
 * supplied was somehow invalid.
 */
char* new_str(const char *str) {
	static char **str_strg;
	static int num_strs;
	char *ret = (char*)0;
	
	/* str was passed; store string on heap */
	if( str ) {
		/* Memory allocations */
		str_strg = (char**)usbs_realloc(str_strg, (num_strs + 1) * sizeof(char*));
		str_strg[num_strs] = (char*)usbs_malloc(strlen(str) + 1);
		/* Copy string */
		strcpy(str_strg[num_strs], str);
		ret = *(str_strg + num_strs);
		++num_strs;
	} 
	/* NULL was passed; do cleanups */
	else if( num_strs > 0 ) {
		while( num_strs > 0 ) {
			free(str_strg[--num_strs]);	
		}
		free(str_strg);
		str_strg = (char**)0;
	}
	
	return ret;
}

/**
 * Finds and writes infotmation for a user to the buffer supplied.
 * uid of user must be given.
 * This function uses /etc/passwd to parse and find the information.
 * Returns a reference to the buffer (null terminated) on success, and
 * and a reference to the *buffer=0 on failure.
 */
char* get_user(enum user_information ui, char *buffer, uid_t uid) {
#define _ETC_PASSWD_FILE   "/etc/passwd"
#define _ETC_PASSWD_DELIM  ':'
/* pos. 3 = uid */
#define _UID_POSITION      3

	char line[LINE_BYTES], uid_buf[BUF_MAX_LEN];
	FILE *passwd;
	
	*buffer = 0;
	
	if( (passwd =  open_file_stream(_ETC_PASSWD_FILE, "r", 0)) < 0 ) goto exit;
	
	/* Go through passwd file */
	while( usbs_fgets((void*)line, sizeof(line), passwd, FALSE) != NULL ) {
		/* Check if line belongs to user given by uid */
		if( atoi(token_at_r(uid_buf, _ETC_PASSWD_DELIM, line, _UID_POSITION)) == (int)uid ) {
			/* Retrieve home directory of user */
			if( !token_at_r(buffer, _ETC_PASSWD_DELIM, line, ui) ) {
				error_out(ERROR_WARNING_ETCPSWD, 0, FLN, (int)uid, _ETC_PASSWD_FILE);
				goto exit;
			} else {
				verbose(VERBOSE_HIGH, "User information (uid=%d) found:\n\"%s\"",
				(int)uid, buffer);
				break;	
			}
		}	
	}
	
	fclose(passwd);
	exit:
	return buffer;
}

/** Block spacing:
 *  __out()   is called at any normal output.
 *  __block() gets called if a block of text is supposed to be sperated
 *            by two \n characters
 *  __init()  is called at application start.
 *  Using these methods simply avoids using two much vertical space on
 *  two consecuetive blocks without using fancy tricks with termcaps.
 */

static BOOL last_is_block;

void __out() {
	last_is_block = FALSE;
}

void __block() {
	if( quiet ) return;
	if( !last_is_block ) putchar('\n');
	last_is_block = TRUE;
}

void __init() {
	last_is_block = TRUE;
}

/**** end block *****/

/**
 * Signal handling: 
 * This variable is for receiving SIGINT and SIGQUIT.
 * The signalhandler for SIGQUIT and SIGINT will check critical.
 * If a critical section is entered, call __start_critical(), if exited
 * call __end_critical()
 * If the program is in a critical section this variable will be set
 * to TRUE. The signal handlers wait until it's falls and will exit the
 * program properly.
 */
static SIG_BOOL require_exit;

/**
 * Signal handler for SIGINT and SIQUIT
 */
RETSIGTYPE quit_signal_handler(int sig) {
	const char *verb = "\n\nReceived SIGINT/SIGQUIT, waiting for save exit time\n\n\0";
	if( write(fileno(stderr), verb, strlen(verb)) < 0 );
	require_exit = TRUE;
}

/** 
 * Is called frequently in main loops of the usbsync program to check
 */

void __start_critical() {
	verbose(VERBOSE_HIGH, " ** Entering critical code section");
	require_exit = FALSE;
}

void __end_critical() {
	verbose(VERBOSE_HIGH, " ** Exitting critical code section");
	if(require_exit) exit(EXIT_SUCCESS);
}

/**
 * Initialized signal handlers for SIGINT and
 * SIGQUIT.
 */ 
void __init_sighandlers() {
	struct sigaction sa;
	sigset_t masked;
	
	memset(&sa, 0, sizeof(struct sigaction));
	memset(&masked, 0, sizeof(sigset_t));
	
	/* Masked while executing sighandler */
	sigemptyset(&masked);
	sigaddset(&masked, SIGINT);
	sigaddset(&masked, SIGQUIT);
	
	/* Set sigaction struct for SIGINT and SIGQUIT */
	sa.sa_handler = &quit_signal_handler;
	sa.sa_mask = masked;
	
	/* Set sigaction for signals SIGINT and SIGQUIT */
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
}

/***** End sighandler routines *****/


/**
 * Draws a visual "seperator" (essentially a line) with sepchar given
 * (NOTE: __init_termcap() must have been called in advance -- redundant see HAVE_TERMCAP)
 */
void draw_separator(char sepchar) {
#define _SEPERATOR_WIDTH 0.8 /* Width of screen seperators like --- or === */
	int i;
	
	/* If option quiet set do not draw anything */
	if( quiet ) return;
	
	for(i = cols*_SEPERATOR_WIDTH; i-->0; ) putchar(sepchar);
	putchar('\n');
}

/**
 * Assemble shortopts string from longopts (struct option).
 * Shortopts string is copied to shortopts_buf.
 * Returns 0 on success, -1 on failure.
 */
int make_shortops(char *shortopts_buf, struct option *lopts) {
	char *p;
	for(p = shortopts_buf; lopts->name != NULL; lopts++) {
			  if( !shortopts_buf ) { 
			  		fprintf(stderr, "Error assembling shortopts %s:%d", 
			  		FLN);
			  		return -1;
			  }
              *p++ = lopts->val;
              if( lopts->has_arg == required_argument || 
                  lopts->has_arg == optional_argument )
                     *p++ = ':';
    }
    *p = '\0';	
    return 0;
}

/**
 * Just a wrapper function to print status messages using the 
 * status.h module.
 */
void wrp_status_print(const char *msg) {
#ifdef HAVE_TERMCAP
	STATUS s = { .term_stream = stdout };
	if( quiet ) return;
	status_print(&s, msg);
#endif
}


char* usbs_cwd(char *buf, size_t len) {
#ifdef HAVE_GETCWD
	return getcwd(buf, len);
#else
#  ifdef HAVE_GETENV
	return getenv("PWD");
#  else
	return NULL;
#  endif
#endif
}

/**
 * Utalizes the terminfo database to perform cursor motion.
 * Returns 0 on success, -1 on failure.
 */
int move_cursor(CURSOR_MOTION motion, int coord) {
#ifdef HAVE_TERMCAP
	char *mot, *pmot;
	int para = -1, i;
	
	/* Retrieve termcap/terminfo */
	__init_termcap();
	
	switch(motion) {
		case UP:
			mot = tigetstr("cuu1");
		break;
		case DOWN:
			mot = tigetstr("cud1");
		break;
		case RIGHT:
			mot = tigetstr("cuf1");
		break;
		case LEFT:
			mot = tigetstr("cub1");
		break;
		case ABS_X:
			mot = tigetstr("hpa");
			para = coord;
		break;
		case ABS_Y:
			mot = tigetstr("vpa");
			para = coord;
		break;
		case CLR_LE:
			mot = tigetstr("el");
		break;
		case CLR_LB:
			mot = tigetstr("el1");
		break;
		/* complete line clearing */
		case CLR_CMPLT:
			if( move_cursor(CLR_LB, 0) < 0 ) return -1;
			if( move_cursor(CLR_LE, 0) < 0 ) return -1;
			return 0;
		break;
		default:
			fprintf(stderr, "Error: Invalid cursor mation\n");
			return -1;
		break;
	}
	
	if( mot <= (char*)0 ) {
		fprintf(stderr, "Error: Cannot retrieve termincal capability\n");
		return -1;
	}
	
	/* If motion has a paramter, perform parameterizing here */
	if( para > -1 ) {
		if( (pmot = tparm(mot, para)) == NULL ) {
			fprintf(stderr, "Error: Cursor motion failed\n");
			return -1;
		}
		i = 1;
	} else {
		pmot = mot;
		i = coord > 0 ? coord : 1;	
	}
	
	/* For terminfo strings that do not have a parameter simply repeat action
	 * coord times. If coord is 0 repeat it only once.
	 */
	
	while(i-- > 0) 
		/* Apply padding and output */
		tputs(pmot, 1, &putchar);
#endif
	return 0;
}

/* Terminal columns and lines; set in __init_termcap() */
int cols;
int lins;

/**
 * Called by each function that uses terminfo.
 * Is used to try to avoid calling setupterm more than once.
 * Returns -1 on failure 0 on success.
 */
int __init_termcap() {
	static BOOL term_initialized;
#ifdef HAVE_TERMCAP
	if( !term_initialized ) {
		/* Retrieve termcap/terminfo */
		if( setupterm((char*)0, fileno(stdout), (int*)0) == ERR )
			return -1; 
		cols = columns;
		lins = lines;
		/* (int*)0 means exit on error) */
		++term_initialized;
	}
#else
#  ifdef HAVE_GETENV
	/* Variables will have been declared by usbsync */
	char *p;
	if( (p = getenv("COLUMNS")) != NULL )
		cols = atoi(p);
	else
#  endif
		cols = 80; /* Default */
#  ifdef HAVE_GETENV	
	if( (p = getenv("LINES")) != NULL )
		lins = atoi(p);
	else 
#  endif
		lins = 25; /* Default */
	term_initialized = TRUE;
#endif
	return 0;
}

// DO REAL GETCHAR

/**
 * Uses the termios to 
 * 
 * If called the first time, this function sets up the terminal so that 
 * no chars will be echoed, carriage return is no new line char anymore
 * and the terminal enters raw mode.
 * When called the second the, the previous terminal settings will be
 * properly restored.
 */
static int raw_term() {
#ifdef HAVE_TERMIOS_H
	static BOOL raw_mode;
	struct termios new_term;
	static struct termios old_term;
	
	if( !raw_mode ) {
		/* Switch off canonical mode (go in raw mode), echoing, and carriage returns as \n */
		if( tcgetattr(fileno(stdin), &new_term) < 0 ) 
			return -1;
		memcpy(&old_term, &new_term, sizeof new_term);
		new_term.c_lflag &= ~ICANON;
		new_term.c_lflag &= ~ECHO;
		new_term.c_lflag |= IGNCR;
		if( tcsetattr(fileno(stdin), TCSANOW, &new_term) < 0 ) 
			return -1;
		++raw_mode;
	} else {
		--raw_mode;
		/* Restore old terminal settings */
		if( tcsetattr(fileno(stdin), TCSANOW, &old_term) < 0 )
			return -1;
		memset(&old_term, 0, sizeof old_term);
	}
#endif
	return 0;
}

/**
 * Like getchar, only cares for that really only one random key 
 * has to be pressed.
 */
int usbs_getchar() {
	int c;
	raw_term();
	c = getchar();
	raw_term();
	return c;
}

/**
 * Prompts a question. Returns the character from options which was chosen by user.
 * Is given a question string and a printf-like printing
 * function for displaying the prompt as well as a string of option characters.
 * The character returned is case in-sensitve. It is always uppercase.
 * Return (char)0 on error.
 */
char prompt_question(const char *qst, const char *options, int (*prnt)(const char *frm, ...)) {
	char q, *hpa, *o, *o2, option_str[32];
	
	/* Assemble options prompt string */
	if( !*options ) return (char)0;
	else {
		o  = (char*)options;
		o2 = option_str;
		while(*o) {
			*o2++ = toupper(*o++);
			*o2++ = '/';
		}
		*--o2 = '\0';
	}
	
	/* Terminfo stuff */
#ifdef HAVE_TERMCAP
	__init_termcap();
	if( !(hpa = tparm(tigetstr("hpa"), 0)) ) return (char)0;	
#endif
	
	/* Enter raw terminal mode */
	raw_term();

	/* Prompt question */
	do {
		prnt("%s (%s):", qst, option_str); 
		q = toupper(fgetc(stdin));
#ifdef HAVE_TERMCAP
		/* Move cursor to the beginning of line, apply padding, output formatted string */
		tputs(hpa, 1, &putchar);
#endif
	} while( !strchr(options, q) );

	/* Exit raw terminal mode */
	raw_term();

#ifdef HAVE_TERMCAP
	/* Clear line */
	move_cursor(CLR_LB, 0), move_cursor(CLR_LE, 0);
#endif 
	
	return q;
}

/**
 * Obtain hostname from /etc/hostname
 * and write to buf.
 */
void hostname(char *buf) {
	int bytes_read;
	int fd;
	
	*buf = 0;
	
	if( gethostname(buf, MAX_HOSTNAME) < 0 ) {
		/* Alternative way; read /etc/hostname */
		if( (fd = open_file_descriptor("/etc/hostname", O_RDONLY, 0444, TRUE)) ) {
			if( (bytes_read = read(fd, buf, 128)) == -1 ) 
			usbs_error(TRUE, errno, FLN, "Unable to read \"/etc/hostname\"");
			buf[bytes_read-1] = '\0';
			close(fd);	
		}
	} 
	 
	/* Fatal error */
	if( !*buf )
		usbs_error(TRUE, errno, FLN, "Unable to obtain host name");
}

/**
 * Write username of currently exectuing user of this process to buffer.
 */
void username(char *buf) {
		get_user(USER_NAME, buf, geteuid());
}


/**
 * Write at string of the form:
 * (user@host:01/05/2009-15:32)
 * to buffer passed. 
 * *buf will be \0 on if something goes wrong, I think ..
 * (The parenthesis are added for formatting convenience)
 */
void usr_host_info(char *buf) {
	char tmp[128];
	struct timeval tv;
	struct tm *t; 
	
	*buf = 0;
	gettimeofday(&tv, NULL);
	t = gmtime(&tv.tv_sec);
	strftime(tmp, sizeof(tmp), "%D-%T", t);
	sprintf(buf, "(%s@%s:%s)", this_username, this_hostname, tmp);
}

/**
 * Converts a numbe of bytes given into a human
 * readable format. Writes it to the buffer *buf.
 */
void bytes_to_human(char *buf, unsigned long int bytes) {
#define KILO_BYTE 1024
	unsigned long int human, tmp;
	char unit[16];
	
	/* bytes */
	if( (tmp = bytes / KILO_BYTE) < 1 ) {
		human = bytes;
		strcpy(unit, "b");
	/* kilo bytes */
	} else if( (human = tmp / KILO_BYTE) < 1 ) {
		human = tmp;
		strcpy(unit, "kb");
	/* mega bytes */
	} else if( (tmp = human / KILO_BYTE) < 1 ) {
		strcpy(unit, "Mb");
	/* giga bytes */	
	} else if( (human = tmp / KILO_BYTE) < 1 ) {
		human = tmp;
		strcpy(unit, "Gb");
	}
	
	sprintf(buf, "%ld %s", human, unit);
}

/**
 * This function is like fprintfs only that it additionally 
 * takes a prefix string to put in fron of every line and that it
 * takes an arguments array instead of a variable argument list (See arg/args.c).
 * Morever, the newline character '\n' always yields a newline with leading 
 * prefix. The vertical tab character '\v' is used for definite new line.
 */
void prfx_print_args(FILE *stream, const char *prfx, const char *frm, ARG_ARRAY args) {
       char *str, *format_p, format[16], format_buf[16];
       BOOL p = FALSE;
       void *next_arg;
       
       format_p = format_buf;
       str = (char *)frm;
       bzero(format_buf, sizeof(format_buf));
       fprintf(stream, "%s: ", prfx);
       for( ; *str; str++ ) {
       		if(*str == '\n') fprintf(stream, "\n%s: ", prfx);
       		else if(*str == '\v') { 
       			/* Use vertical tab '\v' as our new line */
				fputc('\n', stream);
			}
       		else if(*str == '%') {
       			p = TRUE;
       		} else if(p) {
       				/* Check if next argument is available */
       				if( !*args ) continue;
       				if( strchr("sdfhc", *str) != NULL ) {
       					*++format_p='\0';
       					sprintf(format, "%%%s%c", format_buf, *str);
       				}
       				/* Retrieve next argument */
       				if( (next_arg = (*args)->argument) == NULL ) continue;
       			switch(*str) {
       				/* Advance argument array */
#define ARG_OUT p = FALSE, ++args
       				case 'f':
       					fprintf(stream, format, *(double*)next_arg);
       					ARG_OUT;
       				break;
       				case 's':
       					fprintf(stream, format, (char*)next_arg);
       					ARG_OUT;
       				break;
       				case 'd':
       					fprintf(stream, format, *(int*)next_arg);
       					ARG_OUT;
       				break;
       				case 'h':
       					fprintf(stream,"%x", *(unsigned int*)next_arg);
       					ARG_OUT;;
       				break;
       				case 'c':
       					fprintf(stream, "%c", *(char*)next_arg);
       					ARG_OUT;
       				break;
       				case 'p':
       					fprintf(stream, "%p", (void*)next_arg);
       					ARG_OUT;
       				break;
       				default:
       					*format_p++ = *str;
       				break;
       			}
       			if( !p ) {
       				format_p = format_buf;
				}
			} else {
				fputc(*str, stream);
			}  	
	   } 
}

/**
 * Printf like function that takes a prefix.
 * See function above.
 */
void prfx_print(FILE *stream, const char *prfx, const char *frm, va_list ap) {
	ARG_ARRAY args;
	
	/* Convert variadic argument list to argument array */
	args = varg_to_argarr(frm, ap);
	
	/* Call printing function */
	prfx_print_args(stdout, prfx, frm, args);

	/* Free argument array and argument list*/
	arg_array_destroy(args);

	/* Free string storage */
	new_str(0);
}


/**
 * Like prfx_print. See above.
 */
void nv_prfx_print(FILE *stream, const char *prfx, const char *frm, ...) {
	va_list v;
	va_start(v, frm);
	prfx_print(stream, prfx, frm, v);
	va_end(v);
}

/**
 * Does pretty much the same as strdup.
 * Returns a copy of string with new memory alloc'd.
 * Needs to be free'd.
 */
char *cpy_string(const char *str) {
	char *strn = usbs_malloc(strlen(str)+1); /* strlen bytes + \0 */
	strncpy(strn, str, strlen(str));
	strn[strlen(str)]='\0';
	return strn;
}

/**
 * General usbsync memory allocation routine.
 */
void* usbs_malloc(size_t size) {
	void *space;
	if( (space = malloc(size)) == NULL )
		error_out(ERROR_FATAL_MALLOC, errno, FLN);
	return space;
}

/*
 * if realloc fails it returns NULL and leaves everything unchanged
 */
void* usbs_realloc(void *ptr, size_t size) {
	void *space;
	if( (space = realloc(ptr, size)) == NULL )
		error_out(ERROR_FATAL_REALLOC, errno, FLN);
	return space;
}

/**
 * Reallocates memory for an array of nmemb elements of size bytes
 * each. New memory will be initialized to 0.
 * errno will be set by subsequent calls to malloc and realloc
 * if an error occurs.
 */
void* crealloc(void *ptr, size_t nmemb, size_t esize) {
	size_t orig = 0, n = nmemb * esize;
	void *p;
	
	/* FIXME: This appears to be a bug; size of ptr cannot be known! */
	if( ptr ) orig = sizeof ptr;
	
	/* Allocate new memory; make larger */
	if( orig < n ) {
		p = realloc(ptr, n);
		memset( p, 0, n );
	} else if ( n < orig ) {
	/* Free memory; make smaller */
	}
	
	return p;
}

/**
 * Invokes crealloc and adds some error handling to it.
 */
void* usbs_crealloc(void *ptr, size_t members, size_t membsize) {
	void *mem;
	if( (mem = crealloc(ptr, members, membsize)) == NULL ) 
		error_out(ERROR_FATAL_CRLLOC, errno, FLN);	
	return mem;
}

/**
 * General usbsync output function.
 * Understands printf-like syntax (see prfx_print for further details).
 * Always returns 0;
 */
int msg(const char *msg, ...) {
	va_list ap;
	va_start(ap, msg);
	if( strlen(msg) < 1 ) return 0;
	if(!quiet) __out(), prfx_print(stdout, progname, msg, ap);
	va_end(ap);
	return 0;
}

/** 
 * Advances pointer p until character c is encountered in str.
 * p initially points somewhere to a character in str. Returns
 * a reference to p. p might be '\0' in the end.
 */
char* advance_until(char *p, unsigned short c) {
	while(*++p != c && *p);
	return p;
}

/**
 * Parses the nth token from string where each part is delimited by 
 * the delim character.
 * Returns NULL on failure, otherwise a pointer to the token buffer
 * (which is supplied) will be returned.
 * DEPRECATED AND UNDSAFE; DO NOT USE; USE token_at_r() instead
 */
char* token_at(char *token, short delim, const char *str, int n) {
#ifdef HAVE_STRTOK
	char string[LINE_BYTES], d[2] = {delim, '\0'}, *p = string;
	int i = n;
	if( !*str || !strchr(str, delim) ) return NULL;
	bzero(string, LINE_BYTES);
	strcpy(string, str);
	strtok(string, d);
	while( --i > 0 && *p ) p = strtok(NULL, d);
	strcpy(token, p);
#else
	*token = 0;
#endif
	return token;
}

/**
 * Almost the same as without _r. Simply does not use strtok in the implentation
 * and still counts two successive delimiter that would have been missed by
 * the other version. Is also saver than other version. If there is no delimeter
 * to be found returns string as token. If an empty token is encountered at 
 * position n a reference to a null character in token is returned.
 */

char *token_at_r(char *token, unsigned short delim, const char *string, int n) {
	char *start, *end;
	int i, j; 
	
	start = end = (char*)string;
	i = j = n;
	*token = 0;
	
	if( !*string || !string || strcnt(delim, string) < i ) return token;
	
	while( i-- > 0 ) 
		end = advance_until(end, delim);
	while( --j > 0 ) 
		start = advance_until(start, delim);
	
	if( start > string || *start == delim ) ++start;
	
	strncpy(token, start, end - start);
	
	token[end - start] = 0;
	
	return token;
} 

/**
 * Not implemented yet.
 */
char* strchrstr(const char *str, const char *chrs) {
	return (char*)0;
}

/**
 * Obtain user id and gid from /etc/passwd and return a copy of a
 * USER_INFO structure.
 * Return -1 on failing.
 */
USER_INFO getuserinfo(const char *user) {
	char line[512], buf[32]; /* FIXME: line buffer length ok? */
	USER_INFO uinfo;
	FILE *f;
	f = open_file_stream("/etc/passwd", "r", FALSE);
	uinfo.uid = -1, uinfo.gid = -1;
	if( f == NULL ) {
		uinfo.uid = geteuid(), uinfo.gid = getegid();
		usbs_error(FALSE, 0, NULL, 0, "Assuming uid=%d, gid=%d for file permissions\n"
		"(Taken from this process)", uinfo.uid, uinfo.gid);
	} else {
		/* Parse passwd for uid and gid of user */
		while( usbs_fgets(line, 512, f, FALSE) != NULL ) {
			*(line+strlen(line)-1) = '\0';
			if( strncmp(line, user, strlen(user)) == 0) {
					uinfo.uid = atoi( token_at_r(buf, ':', line, USER_ID) );
					uinfo.gid = atoi( token_at_r(buf, ':', line, GROUP_ID) );
					verbose(VERBOSE_MID, "%s: uid=%d, gid=%d", user, uinfo.uid, uinfo.gid);
			}
		}
	}
	if( uinfo.uid < 0 ) {
		uinfo.uid = geteuid(), uinfo.gid = getegid();
		usbs_error(FALSE, 0, NULL, 0, "User \"%s\" not in /etc/passwd, assuming uid=%d, gid=%d\n"
		"for file permissons\n(Taken from this process)", user, uinfo.uid, uinfo.gid);
	}
	if( f != NULL) fclose(f);
	return uinfo;	
}

/**
 * Reverses string *str passed.
 */
void str_reverse(char *str) {
	char *a = str, *b = str + strlen(str) - 1;
	char tmp;
	
	if( !*str ) return;
	
	while( a < b ) {
		tmp = *a;
		*a++ = *b;
		*b-- = tmp;	
	}
}

/** 
 * Append a character to a string.
 */
char* chrcat(char *const dest, const char app) {
	size_t len = strlen(dest);
	dest[len] = app;
	dest[++len] = 0;	
	return dest;
}

/**
 * Remove last character from string
 */
char* last_chr(char *const str) {
	str[strlen(str)-1] = 0;
	return str;
}

/**
 * Like strncat.
 * Only returns (char*)0 if 
 * resulting string would be larger than or equal to len.
 * This is mainly to prevent buffer overflows.
 */
char *usbs_strcat(char *dest, const char *src, size_t len) {
	if( strlen(dest) + strlen(src) + 2 >= len ) {
		return NULL;
	} else 
	return strncat(dest, src, len);
}

/**
 * Return pointer to the first occurrence (in 'str') of
 * any of the characters given in 'chrs' or return pointer to how far
 * we have gotten if none could be found. Additionally, it is an
 * error if character 'reject' is encountered on the way. Pointer
 * in str to how far we have gotten is returned too in this case.
 */
char *strchrstrrjct(const char *str, const char *chrs, char reject) { 
	char *str_p = (char*)str, *chr_p = (char*)chrs;
	for( ; *str_p; str_p++, chr_p = (char*)chrs ) {
		if( *str_p == reject ) return str_p;
		while( *chr_p ) if( *chr_p++ == *str_p ) return str_p;
	}
	return str_p;
}

/**
 * Count character c in str
 */
int strcnt(char c, const char *str) {
	int cnt = 0;
	if( !str ) goto end;
	for( ; *str; str++ )
		if( *str == c ) cnt++;
	end:
	return cnt;
}

/**
 * Like strcnt, only that characters in chrs are counted.
 */
int strmcnt(const char *buf, const char *chr) {
	char *s = (char *)buf;
	int cnt = 0;
	
	if( !buf || !*buf ) goto end; 
	
	do {
		if( strchr(chr, *s) ) cnt++;
	} while(*++s);
	
	end:
	return cnt;
}

/**
 * Counts, in a string whose components are separated by 'char delim', how
 * many tokens, that is, how many gaps between the string delimters, are present.
 */
int gap_count(const char *str, char delim) {
	int delims = 0;
	size_t len;
	
	if( !str ) return 0;
	
	len = strlen(str) - 2;
	
	/* Skip leadin delimeter */
	if( *str == delim ) ++str;
	
	while( *str <= str[len] ) 
		if( *str++ == delim ) 
			++delims;
	
	return delims + 1;
}

/**
 * Counts, in a string whose entries are delimited by 'char delim', the number
 * of tokens that match the string 'token'.
 */
int token_count(const char *str, const char *token, char delim) {
	int i, tok_cnt;
	
	char tok_at_buf[NAME_LENGTH]; /* FIXME: Buffer large enough? */
	
	*tok_at_buf = 0;
	
	if( !str || !token ) return 0;
	
	for( i = 1, tok_cnt = 0; *token_at_r(tok_at_buf, delim, str, i); i++) {
		/* Compare token at position i with token passed in function call */
		if( strcmp(token, tok_at_buf) == 0 ) 
			tok_cnt++;
	}
	
	return tok_cnt;	
}

/**
 * Count printf-like paremters in string.
 * This means paramters like `%s`, `%d` that start
 * with a percent. Only two successive `%` signs
 * are not counted.
 */
int param_cnt(const char *frm) {
	
	char *p = (char*)frm;
	int cnt = 0;
	
	if( !p ) goto end;
	
	for( ; *p; p++ )
		if( *p == *p++  && *p == PARAM_INDICATOR ) 
			cnt++;
	
	end:
	return cnt;
}

static void reverse(char s[]) {
       int c, i, j;
       for (i = 0, j = strlen(s)-1; i < j; i++, j--) {
              c = s[i];
              s[i] = s[j];
              s[j] = c;
       }
}

/* itoa: convert n to characters in s */
void itoa(int n, char s[]) {
       int i, sign;
       if ((sign = n) < 0) /* record sign */
              n = -n;          /* make n positive */
       i = 0;
       do {      /* generate digits in reverse order */
              s[i++] = n % 10 + '0'; /* get next digit */
       } while ((n /= 10) > 0);    /* delete it */
       if (sign < 0)
              s[i++] = '-';
       s[i] = '\0';
       reverse(s);
}
