/*
 *      common.h - part of the usbsync utility
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */
#ifndef COMMON_H
#define COMMON_H 1

/* We are the usbsync program */
#ifndef USBSYNC
#  define USBSYNC 1
#endif

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#else 
#  include <string.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <signal.h>
#include <limits.h>

#ifdef HAVE_UNISTD_H
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <unistd.h>
#endif

/* NOTE: string.h need not to be included in usbsync source files 
         This code will take care of this.
 */
#ifdef STDC_HEADERS
#  include <string.h>
#else
#  ifndef HAVE_STRCHR
#    define strchr index
#  endif
char *strchr ();
#  ifndef HAVE_MEMCPY
#    define memcpy(d, s, n) bcopy ((s), (d), (n))
#  endif
#  ifndef HAVE_STRLEN
size_t strlen(const char *);
#  endif 
#endif

/* Some comptability macros */
#ifndef HAVE_BZERO
#  define bzero(b, len) (memset((b), '\0', (len)), (void)0)
#endif

/* Include headers for time functions */
#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  ifdef HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif

/* Terminal length and height */
extern int cols;
extern int lins;

/* Include general compile-time settings. */
#include "config.h"

/* Working with argument lists */
#include "args/args.h"

/* Booleans */
#define BOOL     int
#define SIG_BOOL volatile sig_atomic_t
#define TRUE     1
#define FALSE    0

/* General buffer lengths */
#define PATH_LENGTH 4096  /* See limits.h; PATH_MAX is unreliable; use own def here */
#define NAME_LENGTH (NAME_MAX + 1)

#define LINE_BYTES   512
#define BUF_MAX_LEN  512
#define USR_MAX_LEN  LOGIN_NAME_MAX /* See limits.h */

/* Maximum visual ubsync filename length, see function FN_(...) and PN_(...) more information. 
 * If termcap is available, these are set dynamically or read from the environemnt. 
 */
#define MAX_FNAME_LEN_L  15 
#define MAX_PATHN_LEN_L  60
#define MAX_FSUFFIX_LEN  10
/* Maximum hostname length */ 
#ifdef _POSIX_HOST_NAME_MAX
#  define MAX_HOSTNAME   _POSIX_HOST_NAME_MAX
#elif HOST_NAME_MAX
#  define MAX_HOSTNAME   HOST_NAME_MAX
#else
#  define MAX_HOSTNAME   128
#endif

/* Miscellenous macros */
#define MASS_STORAGE_NAME   "usb"
#define BACKUPFILE_SUFFIX   ".ubs"
#define EY                  "\0"           /* Empty string */
#define PATH_NAME_DELIM      '/'           /* System's pathname delimiter */
#define PATH_HOME_INIDC      '~'           /* Symbol for home directory */
#define PARAM_INDICATOR      '%'		   /* Specifier for printf-like format */
#define DAY_SECONDS           86400        /* Seconds in a day */

/* Object size macro */
#define SIZE_OF(OBJ) ( sizeof(OBJ) / sizeof(*OBJ) )

/* Macro for calling a destruction function */
#define DESTROY(DESTR_FUNC, OBJ) if(OBJ != NULL) DESTR_FUNC(&OBJ), OBJ = NULL

/* Macro for not-nulling checking and freeing */
#define NFREE(P) if( P )  free(P);

/* Usbsync error stacks.
 * See error/usbs_error.c and error/error.c for more detailed information
 * on error stacks.
 */
enum {
	ESTACK_ERRORS = 0,
	ESTACK_WARNINGS = 1,
};

/* Cursor motion used by function move_cursor */
typedef enum {
	UP, 
	DOWN, 
	LEFT, 
	RIGHT, 
	ABS_X, 
	ABS_Y, 
	CLR_LE,   /* Clear left from cursor */ 
	CLR_LB,   /* Right from cursor */
	CLR_CMPLT /* Complete line clear */
} CURSOR_MOTION;

/*
 * Enum type determines the direction in which files go.
 * Needed by usbs_chown as well as expand_stack (sync.c)
 */
enum expand_type {USB_TO_HD, HD_TO_USB};

/* 
 * User information enum for get_user(...) function.
 * The values of the enum vals specify the position of the
 * corresponding token within the `/etc/passwd' file for
 * a certain entry (token is ':').
 */
enum user_information {
	/* pos. 1 = username */
	USER_NAME = 1, 
	/* pos. 3 = user id */
	USER_ID   = 3,
	/* pos. 4 = group id */
	GROUP_ID  = 4,
	/* pos. 6 = home */
	HOME_DIR  = 6, 
};

/* Program information */
extern const char *progname;
extern const char *version;
extern const char *license;
extern const char *descr;
extern const char *author;
extern const char *year;

/* Directing globals */
extern sig_atomic_t verbosity;
extern int dir_recursion;
extern BOOL quiet;
extern char this_hostname[HOST_NAME_MAX];
extern char this_username[LOGIN_NAME_MAX];
extern char usbsyncfile_name[PATH_LENGTH];

/* Structure to describe a user info */
typedef struct {
	int uid;
	int gid;
} USER_INFO;

/* Prototypes */
extern USER_INFO getuserinfo(const char *user);
extern int strrncmp(const char *str1, const char *str2, size_t n);
extern int __init_termcap();
extern void* crealloc(void *ptr, size_t nmemb, size_t esize);
extern void* usbs_crealloc(void *ptr, size_t members, size_t membsize);
extern void prfx_print(FILE* stream, const char *prfx, const char *frm, va_list ap);
extern void nv_prfx_print(FILE *stream, const char *prfx, const char *frm, ...);
extern void bytes_to_human(char *buf, unsigned long int bytes);
extern void hostname(char *buf);
extern void username(char *buf);
extern void usr_host_info(char *buf);
extern int msg(const char *msg, ...);
extern int make_shortops(char *shortopts_buf, struct option *lopts);
extern char prompt_question(const char *qst, const char * options, int (*prnt)(const char *frm, ...));
extern int move_cursor(CURSOR_MOTION motion, int coord);
extern void draw_separator(char sepchar);
extern void wrp_status_print(const char *msg);
extern char* get_user(enum user_information ui, char *homedir, uid_t uid);
extern char *token_at_r(char *token, unsigned short delim, const char *string, int n);
extern char* PN_(const char *path, BOOL);
extern void __init_path_display();
extern void __deinit_path_display();
extern int strmcnt(const char *buf, const char *chr);
extern char *usbs_strcat(char *dest, const char *src, size_t len);
extern void prfx_print_args(FILE *stream, const char *prfx, const char *frm, ARG_ARRAY args);
extern void str_reverse(char *str);
extern int strcnt(char c, const char *str);
extern int gap_count(const char *str, char delim);
extern int token_count(const char *str, const char *token, char delim);
extern int usbs_getchar();
extern char* chrcat(char *const dest, const char app);
extern char* last_chr(char *const str);

/* Block output functions **/
extern void __out();
extern void __block();
extern void __init();

/* Signal handling functions */
extern void __init_sighandlers();
extern void __start_critical();
extern void __end_critical();

/* Auxiliary */
extern void* usbs_malloc(size_t);
extern void* usbs_realloc(void *ptr, size_t);
extern void itoa(int n, char s[]);
extern char *cpy_string(const char *str);
extern char* new_str(const char *str);
extern char* usbs_cwd(char *buf, size_t len);
 
#endif
