/*
 *      debug.c - Debugging, dumping and synchronization statistics routines
 * 		Part of the usbsync utility 
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "debug.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <libgen.h>
#include <sys/time.h>

#include "../common.h"
#include "../options/options.h"
#include "../file.h"
#include "../args/args.h"
#include "../stack/stack.h"
#include "../error/usbs_errors.h"
#include "../sync.h"

/* Absolute value of number x */
#define ABSLT(x) x < 0 ? -x : x

/* Prototypes */
void dump_stack(STACK *s, enum dump_type stack_data);

/* 
 * Counts how many errors occured.
 * If error_cnt = -1 the program had a fatal interrupt.
 * At program start, must
 */
int error_cnt, warning_cnt;

/* 
 * Synchronization statistics.
 */
int dirs_synced;  			/* how many directories synchronized */
int dirs_check;  		    /* how many dirs checked */
int dirs_skipped; 			/* how many dirs skipped (only with -p and -c) */  
int files_synced;           /* how many files synced */
int files_check;  			/* how many files checked */
int files_skipped; 			/* how many files skipped (only with -p and -c) */
int files_removed;

#define _TIME_SECONDS 1000*1000 	/* 1s = 1000*1000us */
time_t time_needed; 				/* synchronization and checkind time needed*/

/**
 * Initialize dir, file and time statistcs.
 */
void __init_statistics() {
	dirs_synced = dirs_check = dirs_skipped =
	files_synced = files_check = files_skipped = files_removed =   
	time_needed = error_cnt = warning_cnt = 0;
}

/**
 * Called at the end, summarizes how many errors occurred, etc..
 */
void print_conclusion() {
	char 
	dirs_skipped_str[128], 
	files_skipped_str[128], files_removed_str[128], 
	total_skipped[128];
	double time;
	*dirs_skipped_str = *files_skipped_str = *total_skipped = *files_removed_str = '\0';
	
	/* If option quiet set do not output anything */
	if( quiet ) return;
	
	__block();
	putchar('\n');
	draw_separator('-');
	putchar('\n');
	
	/* Error summary */
	switch(error_cnt + warning_cnt) {
		case -1:
			msg("** Program aborted abnormally. Error code: %d\v", last_error_code);
			return;
		break;
		case 0:
			msg("** Synchronization entirely successful\v");
		break;
		default:
#define NOT_SHOWN (_option_check(_OPTION_SUPPR_ERR) ? ("(not shown) ") : (EY))
#define PLURAL(X) ((X > 1 || X == 0) ? ("s") : (EY))
			msg("** There were %d %serror%s and %d %swarning%s.\n"
				"** Double-check your \"%s\" file\n"
				"** as well as any other files involved.\v",
				error_cnt,   NOT_SHOWN, PLURAL(error_cnt), 
				warning_cnt, NOT_SHOWN, PLURAL(warning_cnt), 
				usbsyncfile_name);
		break;
	}	
	
	putchar('\n');
	if( dirs_synced == 0 && 
	 	files_synced == 0 && 
	 	/* error_cnt == 0 && */
	 	files_skipped == 0 &&
	 	dirs_skipped  == 0 &&
	 	files_removed  == 0) {
		msg("** No synchronization tasks have been performed\v\v");
	}
	
	/* Sync statistics */
	 
#define LINE_EXT "---------------"
	 	 if( dirs_skipped  > 0 ) 
	 	 	sprintf(dirs_skipped_str,  "  | skipped: %3d", dirs_skipped);
	 	 if( files_skipped > 0 ) 
	 	 	sprintf(files_skipped_str, "  | skipped: %3d", files_skipped);
	 	 if( dirs_skipped + files_skipped > 0)
	 	 	sprintf(total_skipped, "  | skipped: %3d", files_skipped + dirs_skipped); 
	 	 if( files_removed > 0 ) 
	 	 	sprintf(files_removed_str, "  | removed: %3d", files_removed);
	 	 	
	 	 time = ((double)time_needed)/((double)(_TIME_SECONDS));
	 	 	
		 msg("Synchronization statistics:\n"
			 "  Directories   :  %3d  | checked:  %3d%s\n"
			 "  Files         :  %3d  | checked:  %3d%s%s\n"
			 "  ----------------------------------------%s%s\n"
			 "  Total         :  %3d  | checked:  %3d%s\n"
			 "  Time elapsed  : %#.3fs\v",
			 dirs_synced, dirs_check, dirs_skipped_str,
			 
			 files_synced, files_check, files_skipped_str, files_removed_str,
			 
			 (files_skipped>0||dirs_skipped>0)?LINE_EXT:(files_removed>0?"---":EY),
			 (files_removed>0)?LINE_EXT:EY,
			 
			 dirs_synced+files_synced, files_check+dirs_check, total_skipped,  ABSLT(time));
}

/**
 * Time measurement functions.
 */
static time_t initial_time;

void __start_timer() {
	struct timeval init_tv;
	gettimeofday(&init_tv, NULL);
	initial_time = init_tv.tv_usec;
}

/**
 * Return time since last __start_timer() call in cpu clock ticks
 * Returns -1 if an error occurred.
 */
time_t __stop_timer() {
	struct timeval final_tv;
	time_t time_taken;
	if( initial_time == (time_t)0 ) return (time_t)0;
	gettimeofday(&final_tv, NULL);
	time_taken = final_tv.tv_usec - initial_time;
	initial_time = (time_t)0;
	return time_taken;
}

/*** end time measurement routines ****/


/** Dumping routines */
/** 
 * General dump string
 */
static void dumpstring(enum dump_type type, const char dump_name[], int space, const void *dump_data) {
	int i;
	assert(space > strlen(dump_name));
	printf("%s:     %p: ", progname, (void*)dump_data); 
	printf("(%s)", dump_name);
	for(i=0; i < space-strlen(dump_name); i++) putchar(' ');
	switch(type) {
		case STR:
			printf(": \"%s\"\n", (char*)dump_data);
		break;
		case INT:
			printf(": %d\n", *((int *)dump_data));
		break;
		case HEX:
			printf(": %#08x\n", (unsigned int)dump_data);
		break;
		case BOL:
			printf(": %s\n", *((int *)dump_data)==0?"false":"true");
		break;
		case CHR:
			printf(": %c\n", *(char*)dump_data);
		break;
		case PTR:
			printf(": %p\n", (void*)dump_data);
		break;
	}
}

/**
 * Dump a USBS_FILE structure
 */
void dump_usbs_file(USBS_FILE *f) {
	assert(f != NULL);
	if( !f ) return;
	msg("  usbs_file dump (%p)\v", (void*)f);
	dumpstring(STR, "f->usb_fname",  18, PN_(f->usb_fname, 1));
	dumpstring(STR, "f->host_fname", 18, PN_(f->host_fname, 1));
	dumpstring(INT, "f->uid", 18, &f->uid);	
	dumpstring(INT, "f->gid", 18, &f->gid);
	dumpstring(INT, "f->usb_time", 18, &f->usb_time);
	dumpstring(INT, "f->host_time", 18, &f->host_time);	
	dumpstring(STR, "f->file_type", 18, file_type_names[f->file_type]);
}

/**
 * Dumping a variable argument strucutre.
 * See args/args.h for mor information.
 */
void dump_arg_list(ARG_ARRAY l) {
	struct arg **al = l;
	if( !al ) {
		msg(" Dump not possible, argument list == NULL\v");
		return;	
	}
	
	msg(" ** arg_list_dump DUMP: (0x%h)\v",  al);
	for( ; *al; al++ ) {
		msg(" next argument DUMP:\v");
		dumpstring(CHR, "*a->argument_type", 25, &(*al)->argument_type);
		
		/* Different dumps for different types of arguments */
		switch( (*al)->argument_type ) {
#define ARG_DUMP(X) dumpstring(X, "*a->argument", 25, (*al)->argument);
			case 's':
				ARG_DUMP(STR);
			break;
			case 'd':
				ARG_DUMP(INT);
			break;
			case 'c':
				ARG_DUMP(CHR);
			break;
		}
	}		
	msg(" ** end arg_list dump\v");
}

/**
 * Dumping an error message structure.
 * See error/error.h for more information.
 */
void dump_error_message(ERROR_MESSAGE *e) {
#ifdef DEBUG
	struct error_info *ei;
	assert( e != NULL );
	
	/* Dumping struct ERROR_MESSAGE */
	msg("  error_message dump (%p)\v", e);
	dumpstring(INT, "e->code", 18, &e->code);
	dumpstring(STR, "e->msg",  18, e->msg);
	
	/* Dumping struct error info */
	if( (ei = e->err_info) == NULL ) {
		msg("  -- no error information: (null)\v");
		return;
	}
	msg("    error_info: (0x%h)\v", e->err_info);
	dumpstring(STR, "ei->file",   18, ei->file);
	dumpstring(INT, "ei->line",   18, &ei->line);
	dumpstring(INT, "ei->err_no", 18, &ei->err_no);
	
	/* Dumping struct args */
	if( ei->args == NULL ) {
		msg("  -- no arguments for message: (null)\v");
		return;
	}
	
	dump_arg_list(ei->args);
#endif	 
}

void dump_usbsyncdata(USBS_SYNCDATA *sd) {
	assert( sd != NULL );
	putchar('\n');
	msg("Dumping USBS_SYNCDATA (%p)\n ", sd), draw_separator('-');
	msg("Dumping FILE_STACK:");
	dump_stack(sd->sync_files, 0);
	msg("Dumping PREFIX_IGN_STACK:");
	dump_stack(sd->prefix_ign, STR);
	msg("Dumping SUFFIX_IGN_STACK:");
	dump_stack(sd->suffix_ign, STR);
	msg("FLAGS:\v");
	dumpstring(INT, "sd->flags", 18, &sd->flags);
	msg(" "), draw_separator('-');
}

/**
 * Dumps a stack. See stack/stack.c for more information.
 * Note that the last enum does only apply if stack is of type
 * (stack->stack_type) is set to NORMAL_STACK. Then all data
 * of this stack is dumped as this type.
 */
void dump_stack(STACK *s, enum dump_type stack_data) {
	STACK_ELEMENT *e;

	if( s == NULL ) {
		usbs_error (FALSE, 0, FLN, "Dump not possible (stack==NULL)");
		return;	
	} else {
		__block();
		printf("%s: **** usbsync STACK DUMP (%p):\n", progname, (void*)s);
		dumpstring(INT,"s->count", 18, &s->count);
		dumpstring(HEX,"s->top", 18, s->top);
#ifdef MULTISTACK
		dumpstring(STR,"s->stack_type", 18, stack_type_names[s->stack_type]);
#endif
		for( e = s->top; e != NULL; e = e->next) {
#define NORMAL_DUMP dumpstring(stack_data, "f->data", 18, e->data)
#ifdef MULTISTACK
			/* Different dumps for different types of stacks */ 
			switch(s->stack_type) {
				case NORMAL_STACK:
					NORMAL_DUMP;
				break;
				case FILE_STACK:
					dump_usbs_file((USBS_FILE*)e->data);	
				break;
				case ERROR_STACK:
					dump_error_message((ERROR_MESSAGE*)e->data);
				break;		
				case SYNCDATA_STACK:
					dump_usbsyncdata((USBS_SYNCDATA*)e->data);
				break;
			}
#else
				NORMAL_DUMP;
#endif				
		}	
		msg("**** END DUMP\v");
	}
}

void dump_file_stack(STACK *s) {
	dump_stack(s, HEX);
}

/**
 * Verbose output according to verbosity level
 * given by extern int verbosity (0 < level < 4)
 * 
 */
void verbose(int level, const char *msg, ... ) {
       char prfx[127];
       va_list ap;
       va_start(ap, msg);
       /* level check */
#ifdef _VERBOSE_ALLOWED
       if( level <= verbosity && verbosity > 0 )
#endif
       sprintf(prfx, "%s: verbose[%d]", progname, level), __block(),
       prfx_print(stdout, prfx, msg, ap), putchar('\n'), putchar('\n');
}

BOOL verbosity_check(int level) {
#ifdef _VERBOSE_ALLOWED
      if( level <= verbosity && verbosity > 0 )
      return TRUE;
#endif
      return FALSE;
}



