/*
 *      file.c - General file handling routines
 * 		Part of the usbsync utility 
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */


#define _ATFILE_SOURCE

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#endif

#include "file.h"

#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <utime.h>
#include <unistd.h>
#include <dirent.h>
#include <libgen.h>

#include <fcntl.h>

#include "common.h"
#include "path.h"
#include "debug/debug.h"
#include "options/options.h"
#include "error/usbs_errors.h"

#ifdef HAVE_TERMCAP
#  include "status/status.h"
#endif 

/* Names for file types; important for debugging */
char *file_type_names[] = {
	"NORMAL_FILE",
	"NORMAL_DIR",
	"EMPTY_DIR"
};

USBS_FILE* file_new() {
	USBS_FILE *file = usbs_malloc(sizeof(USBS_FILE));
	file->usb_fname = file->host_fname = NULL;
	file->uid = -1;
	file->gid = -1;
	file->usb_time = 0;
	file->host_time = 0;
	file->file_type = NORMAL_FILE;
	return file;
}

void file_destroy(USBS_FILE **file) {
	assert(file != NULL);
	if( !*file ) return;
	if( (*file)->usb_fname != NULL )
		free((*file)->usb_fname);
	if( (*file)->host_fname != NULL )
		free((*file)->host_fname);
	free(*file);
	*file = NULL;
}

/**
 * 
 * NOTE: Caller needs to fclose(fstream);
 * Returns NULL on failure.
 */
FILE* open_file_stream(const char *file, const char *mode, BOOL exit_on_failure) {
	FILE *fstream;
	if( (fstream = fopen(file, mode)) == NULL )
		error_out(ERROR_ERROR_FOPEN, errno, FLN, PN_(file, 1));
	return fstream;
}

/**
 * Achievies pretty much the opposite of basename. Writes the 
 * directory of the file/directory specified by path to dir.
 */
void get_directory(char *dir, const char *path) {
	char *p;
	*dir = 0;
	check_path_length("General", path, 0);
	strncpy(dir, path, PATH_LENGTH);
	dir[PATH_LENGTH-1] = 0;
	p = dir + strlen(dir);
	while( *--p != PATH_NAME_DELIM && p > dir);
	*p = 0;
}

/**
 * Concatenate two directories or files
 */
void concat_dirs(char *end_dir, const char *d1, const char *d2) {
	char *dir1 = (char*)d1, *dir2 = (char*)d2;
	assert(end_dir != NULL && dir1 != NULL && dir2 != NULL);
	if( strlen(dir1) > 0 && strlen(dir2) > 0 ) {
		/* remove trailing or leading '/' */
 		if( dir1[strlen(dir1)-1] == PATH_NAME_DELIM ) dir1[strlen(dir1)-1] = '\0';
		if( *dir2 == PATH_NAME_DELIM ) ++dir2;
	}
	sprintf(end_dir, "%s%c%s", dir1, PATH_NAME_DELIM, dir2);	
}	

/**
 * Fgets from stream with error check routines.
 */
void* usbs_fgets(void *buf, int size, FILE *fstream, BOOL exit_on_failure) {
	if( ferror(fstream) ) {
		error_out(ERROR_ERROR_FSTREAM, errno, FLN, fstream);
		clearerr(fstream);
		return NULL;
	} else {
		return fgets(buf, size, fstream);
	}
	return NULL;
}


/**
 * Self-explanatory 
 */
static void zero_out(char *str) {
	for( ; *str != '\n' && *str != '\0'; str++ ) *str = '\0';
	*str = 0;
}

/**
 * Read line from stream whereby an allocated memory space is returned.
 * This function does not have any length restrictions to the line being read.
 * The line will contain the '\n' character in the end.
 * It could happend that not enough memory is available, in this case
 * the line is "silently truncated". This should not happen by default.
 * If the string allocated herein is not used anymore this function has to called
 * with the stream argument as NULL in the end to perform cleanups.
 */
char *readline_stream(FILE *s) {
#define LINE_END(C) (C == '\n' || C == '\0') 

	/* We start with a 64<<1 length buffer */
	static char *line_buf;
	size_t i = 0, sz = 0x4;
	
	if( !s ) {
		/* Freeing the buffer */
		if( line_buf ) {
			free(line_buf);
			line_buf = NULL;
		}
		return NULL;
	}
	
	if( line_buf ) zero_out(line_buf);
	
	/* Read line */
	do {
		
		/* Enlarge buffer and allocate new memory */
		sz <<= 1;
		if( ( line_buf = realloc(line_buf, sz * sizeof(char)) ) == NULL ) {
			/* First time must yield memory */
			if( i == 0 ) { 
				/* Will abort */
				error_out(ERROR_FATAL_REALLOC, 0, FLN);
			}
			/* Truncate if not enough memory */
			line_buf[i] = '\n';
		}
		
		/* Read file character by character */
		do {
			if( ferror(s) ) {
				error_out(ERROR_ERROR_FSTREAM, 0, FLN, (void*)s);
				return NULL;
			}
			if( (line_buf[i++] = fgetc(s)) == EOF ) 
				return NULL;
		} while( !LINE_END(line_buf[i-1]) && i+1 < sz ); 
		
	} while( !LINE_END(line_buf[i-1]) );
	
	/* Null terminating */
	line_buf[i] = 0;
	
	return line_buf;
}


/**
 * NOTE: Caller needs to close(fd);
 * Returns -1 on failure.
 */
int open_file_descriptor(const char *file, int flags, mode_t mode, BOOL exit_on_failure) {
	int fd;
	if( (fd = open(file, flags, mode)) < 0 )
		error_out(ERROR_ERROR_FOPEN, errno, FLN, PN_(file, 1));
	return fd;
}

/**
 * Does the same as utime, only does not follow symlinks. 
 * If atime == 0 && mtime == 0 in struct utimbuf the current time will
 * be taken and set for for both.
 */
static int utime_symlink(int atime, int mtime, const char *path) {
#ifdef HAVE_UTIMENSAT
	char current_dir[PATH_LENGTH], path_dir[PATH_LENGTH];
	struct stat st;
	struct timespec sp[2] = { {atime, 1}, {mtime, 1} };
	int ret;
	if( usbs_cwd(current_dir, PATH_LENGTH) == NULL ) {
		error_out(ERROR_ERROR_GETCWD, errno, FLN);
		return -1;	
	} 
	get_directory(path_dir, path);
	chdir(path_dir);
	/* FIX: utimensat seem to be fairly new and buggy (as of kernel 2.6.x or so), 
	 * hence these hacks in here; is there any other more general method? 
	 */
 	ret = utimensat(AT_FDCWD, basename((char*)path), sp, AT_SYMLINK_NOFOLLOW);
 	chdir(current_dir);
 	lstat(path, &st);
 	verbose(VERBOSE_HIGH, "utimensat: cwd=\"%s\"\npath_dir=\"%s\"\nret=%d, atime=%d, "
 	"mtime=%d\nreal_mtime=%d",
 	current_dir, path_dir, ret, atime, mtime, st.st_mtime);
 	return ret;
#else
#  warning utimensat() unknown. Will not be able to set symbolic link modification times.
 	return 0;
#endif
}
 
static int utime_symlink_cnfrm(const char *filename, const struct utimbuf *times) {
	return utime_symlink(times->actime, times->modtime, filename);
}

/**
 * Returns the size of a file in bytes.
 * If path is a symlink, the size of the link's target is returned.
 * If path is a dangling symlink, 0 is returned.
 * If an error occurs, this function exists with -1.
 */
int true_filesize(const char *path) {
	struct stat s;
	if( is_dangling_symlink(path) ) return 0;
	if( stat(path, &s) < 0 ) return -1;
	return s.st_size;
}

/**
 * Writes a formatted information string about the symbolic link given in path.
 * On an error or if no info could be retrieved buffer only contains single '\0'.
 */
void symlink_info(char *buf, const char *descr, const char *path) {
#define NO_INFO strcpy(buf, "No information")
#ifdef HAVE_READLINK
	char rdlnk[PATH_LENGTH];
	*buf = 0;
	bzero(rdlnk, PATH_LENGTH);
	if( !is_symlink(path, 1) || readlink(path, rdlnk, sizeof rdlnk - 1) < 0 ) {
		NO_INFO;
		return;
	}
	rdlnk[strlen(rdlnk)] = '\0';
	sprintf(buf, "%s: \"%s\"", descr, rdlnk);
#else
	NO_INFO;
#endif
}


/**
 * Is passed a path. Returns TRUE if file at path is a symlink.
 * Returns FALSE if not or if an error occurred. Outputs an error message.
 * If silent is set, function will not creat any error message if path
 * to be checked does not exist; it will simply return FALSE.
 */
BOOL is_symlink(const char *path, BOOL silent) {
	struct stat s;
	if( lstat(path, &s) < 0 ) {
		if( !silent )
			error_out(ERROR_ERROR_LSTAT, errno, FLN, path);
		return FALSE;
	}
	return S_ISLNK(s.st_mode) ? TRUE : FALSE;
}

/**
 * A dangling symbolic link is defined as a symbolic link which
 * points to a non-existing other file. Returns TRUE if so, FALSE if 
 * not.
 */
BOOL is_dangling_symlink(const char *path) {
	/* Check if symlink and if non-accessible */
	return is_symlink(path, TRUE) && (access(path, F_OK) != 0);
}

/** 
 * A file exists if it is either a simlink (dangling or not)
 * or if it can be normally accessed anyways.
 */
BOOL chk_existence(const char *path) {
	return is_symlink(path, TRUE) || (access(path, F_OK) == 0);
}

/**
 * Sets the mode of file2 equal to the mode of file1.
 * By mode is meant all the file protection bits as returned by fstat.
 * Files are given by their absolute path.
 * Return -1 on failure and 0 on success.
 */
int mode_equate(enum expand_type direction, const char *file2, const char *file1) {
	struct stat s;
	int (*chmod_func)(const char *, mode_t) = &chmod;
	
	/* 
	FIXME: Redundant - `-o' ownly for chown() by definition
	if( direction == HD_TO_USB && _option_check(_OPTION_IGN_OWNER) ) {
		verbose(VERBOSE_LOW, 
		"Skipping chmod() due to `-o` option being set:\n\"%s\"", file2);
		return 0;
	}
	*/
	
	/* Use different chmod for symbolic links */
	if( _option_check(_OPTION_CPY_SYMLINKS) ) chmod_func = &chmod; 
	/* FIXME: lchmod - there appears to no generally good implementation for lchmod */
	
	/* Get source mode information */
	if( _option_check(_OPTION_CPY_SYMLINKS) ) {
	    if( lstat(file1, &s) < 0 ) {
			error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(file1, 1));
			return -1;
		}
	} else if( lstat(file1, &s) < 0 ) {
		error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(file1, 1));
		return -1;
	}

	/* Set mode bits */	
	if( chmod_func(file2, s.st_mode) < 0 ) {
		error_store(ESTACK_ERRORS, ERROR_ERROR_CHMOD, errno, FLN, PN_(file2, 1));
		return -1;	
	}
	
	verbose(VERBOSE_HIGH, "Setting file mode from source file:\n\"%s\" (mode: %d)\n\"%s\"", 
	PN_(file1, 1), s.st_mode, PN_(file2, 1));
	return 0;
}

/**
 * Return stat information to stat struct supplied.
 * The type of stat routine used is determined by the
 * current run time settings.
 * Returns status of the stat function.
 */
int usbs_stat(char *f, struct stat *s) {
	int status = -1;

#ifndef AC_FUNC_LSTAT_FOLLOWS_SLASHED_SYMLINK
	if( f[strlen(f)] == PATH_NAME_DELIM ) f[strlen(f)] = '\0';
#endif
	
	if( _option_check(_OPTION_CPY_SYMLINKS) ) {
		if( (status = lstat(f, s)) < 0 )
			error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(f, 1));
	} else if( (status = stat(f, s)) < 0 ) {
			error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(f, 1));
	}

	return status;
}

/**
 * If file at path is a symbolic link with a target the target
 * of the symlink is set to have the same atime and mtime as the 
 * symlink itself
 * Return -1 on failure and 0 on success.
 */
static int equate_symlink_target_mtime(const char *path) {
	struct stat s;
	struct utimbuf u;
	
	if( !is_symlink(path, TRUE) ) return -1; 
	if( is_dangling_symlink(path) ) return -1;
	if( lstat(path, &s) < 0 ) return -1;
	u.actime = s.st_atime, u.modtime = s.st_mtime;
	if( utime(path, &u) < 0 ) return -1;
	
	return 0;
}

/**
 * Sets mtime and atime of dest to that of src.
 * If only_symlink is set this means that no utime call other than those
 * for symlinks will be performed.
 * Returns 0 on success, -1 on failure.
 */
int equate_am_time(const char *src, const char *dst, BOOL only_symlink) {
	struct stat src_lstat;   
	struct utimbuf tb;
	
	/* Check for existence of src and dst */
	if( !chk_existence(src) || !chk_existence(dst) ) {
		error_out(ERROR_ERROR_MODTIME, 0, FLN, src, dst);
		return -1;
	}
	
	/* Get source information */
	if( lstat(src, &src_lstat) < 0 ) {
		error_out(ERROR_ERROR_MODTIME, errno, FLN, src);
		return -1;
	}
	
	tb.actime = src_lstat.st_atime, tb.modtime = src_lstat.st_mtime;
	
	/* Src and dst should now exist. And in case of only_symlinks they are both symlinks and regarded
	 * as such. In any other case they are assumed to be normal files.
	 */
	
	/* If only symlink */
	if( only_symlink ) {
		verbose(VERBOSE_MID, "Set modification times of symlinks equal:\nsrc=\"%s\""
		"\ndst=\"%s\"", src, dst);
		if( utime_symlink(src_lstat.st_atime, src_lstat.st_mtime, dst) < 0 )
			return -1;	
		return 0;
	}
	
	/* Now we utime for regular files. */
	
	/* In order to avoid backwards copying we set utime of symlink and real file the same of src */
	if( is_symlink(src, TRUE) && !is_dangling_symlink(src) ) {
		verbose(VERBOSE_HIGH, "Set mod time of symlink and its target equal:\nfile=\"%s\"", src);
		if( utime(src, &tb) < 0 ) {
			error_out(ERROR_ERROR_UTIME, errno, FLN, src);
			/* simply continue if any error occured here */	
		} 
	}
	
	/* Normal utime for dest */
	if( utime(dst, &tb) < 0 ) {
		verbose(VERBOSE_HIGH, "Setting mod times equal:\nsrc=\"%s\""
		"\ndst=\"%s\"", PN_(src, 1), PN_(dst, 1));
		error_out(ERROR_ERROR_UTIME, errno, FLN, dst);
		return -1;
	}
	
	/* If destination was a symlink set the time for this link also (cannot be dangling) */
	if( is_symlink(dst, 1) ) {
		verbose(VERBOSE_MID, "Destination was also symlink, set mod times:\ndst=\"%s\"", dst);
		if( utime_symlink_cnfrm(dst, &tb) < 0 ) {
			error_out(ERROR_ERROR_MTUPDT, errno, FLN, src);
		/* also simply continue if any error occured here */	
		}
	}
	
	verbose(VERBOSE_MID, "mtime=%d, atime=%d set for\nsrc=\"%s\"\ndst=\"%s\"", 
	src_lstat.st_mtime, src_lstat.st_atime, PN_(src, 1), PN_(dst, 1));
	
	return 0;
}

BOOL is_dir_ftype(enum usbs_filetype ftype) {
	return (ftype == NORMAL_DIR || ftype == EMPTY_DIR || ftype == UPDATE_DIR);
}

/**
 * Checks a usbs file and returns true if its some form of directory.
 */
BOOL is_directory(USBS_FILE *f) {
	return is_dir_ftype(f->file_type);
}

/** 
 * Usbsync wrapper function for chown. 
 * Copy direction enum is passed as paramter.
 * This function returns the exit status of 
 * plus performs some additional checks and may
 * give do additional outputs on errors.
 */
int usbs_chown(enum expand_type direction, const char *path, uid_t uid, gid_t gid) {
	
	int result = 0;
	int (*chown_func)(const char *, uid_t, gid_t) = &chown;
	
	/* Use different chown for symbolic links */
	if( _option_check(_OPTION_CPY_SYMLINKS) ) chown_func = &chown; 
	/* FIXME: lchown() - no implementation yet available */
	
	if( direction == HD_TO_USB && _option_check(_OPTION_IGN_OWNER) ) {
		verbose(VERBOSE_LOW, "Skipping chown() due to `-o' option being set:\n\"%s\"", path);
		return 0;
	}
	
	/* Otherwise perform and return normal chown sys call */
	verbose(VERBOSE_MID, "System call chown() performed on:\n\"%s\"", path);
	
	if( (result = chown_func(path, uid, gid)) < 0 ) {
	 	error_store(ESTACK_ERRORS, ERROR_ERROR_USBCHOWN, errno, FLN, 
		path, (int)uid, (int)gid);
	}
	
	return result;
}

/**
 * Copys a file, source to destination. Returns 0 on success. On failure this
 * function will return -1. Errno is set accordingly.
 */
int cpy_file(const char *src, const char *dst, int uid, int gid, BOOL exit_on_failure) {
	int source, destination;
	char line[512], copied_info[512];
	struct stat src_stat;
	int bytes_read, bytes_written, bytes_total_written = 0;
#ifdef HAVE_TERMCAP
	STATUS *s; /* Status percentage display */
#endif
	
	source = open_file_descriptor(src, O_RDONLY, S_IRWXU | S_IRGRP | S_IROTH, FALSE);
	destination = open_file_descriptor(dst, O_WRONLY | O_CREAT | O_TRUNC, 0766, FALSE);
	
	if( source < 0 || destination < 0) return -1;
	
	if( stat(src, &src_stat) < 0 ) {
		error_out(ERROR_ERROR_LSTAT, errno, FLN, src);
		return -1;
	}

#ifdef HAVE_TERMCAP
	if( (s = status_new(stdout, src_stat.st_size, 0)) == NULL ) {
		verbose(VERBOSE_LOW, "Setting up status pecentage display failed");
	} else {
		/* This means we will only see every 10 percent */
#  define STATUS_SKIP_FRACT 20
		s->skip_fraction = src_stat.st_size / sizeof line / STATUS_SKIP_FRACT;
		verbose(VERBOSE_HIGH, "skip_fraction=%d for st_size=%d", 
		s->skip_fraction, src_stat.st_size); 
	}
#endif
	
	/* Read bytes */
	while( (bytes_read = read(source, line, sizeof(line))) > 0 ) {
			
		/* Check if src fd is still valid */
		if( source < 0 ) {
			error_out(ERROR_ERROR_LSTAT, errno, FLN, source);
			return -1;
		}
		
		/* Write bytes */
		if( (bytes_written = write(destination, line, bytes_read)) < bytes_read ) {
			error_out(ERROR_ERROR_BYWRIT, errno, FLN, destination);
			return -1;	
		}
		
		/* Check if dst fd is still valid */
		if( destination < 0 ) {
			error_out(ERROR_ERROR_BYWRIT, errno, FLN, destination);
			return -1;
		}
		
		bytes_total_written += bytes_written;
		
#ifdef HAVE_TERMCAP
		/* Update status if feature is available */
		if( s != NULL && !quiet ) status_update(s, bytes_total_written);
#endif
	}
	
	/* Flush destination fd so that data is sure to have been copied */
	if( fsync(destination) < 0 ) {
		error_out(ERROR_ERROR_FFLUSH, errno, FLN, destination);
		close(source); 
		close(destination);
		return -1;	
	}
	
	if( close(source) < 0 || close(destination) < 0 ) {
		error_out(ERROR_ERROR_FDCLOSE, errno, FLN);
		return -1;	
	}
	
#ifdef HAVE_TERMCAP
	/* Only use status bar if initialization was successful */
	if( s ) {
		bytes_to_human(copied_info, src_stat.st_size);
		if( !quiet ) status_print(s, "[%s]", copied_info);
		status_destroy(s);	
	}
#endif
	
	/* Preserve mtime and atime */
	if( equate_am_time(src, dst, FALSE) < 0 ) {
		error_out(ERROR_ERROR_SETMTM, 0, FLN);
		return -1;
	}
	
	return 0;
}

/**
 * Copy a symbolic link rather than the files where it points to.
 * Return 0 on success, -1 on failure, and 1 if *src is not a symbolic link.
 * Sets owner to uid and gid. If aggressive is set any kind of file
 * will be overwritten for dst.
 */
/* uid and gid are redundant for symlinks */
int cpy_symlink(const char *src, const char *dst, int uid, int gid, BOOL aggressive) {	
	char link_buf[PATH_LENGTH];
	bzero(link_buf, PATH_LENGTH); /* needed because readlink does not \0-terminate */
	
	/* Remove dst file, no matter what */
	if( aggressive && chk_existence(dst) ) {
		if( unlink(dst) < 0 ) 
			error_out(ERROR_ERROR_UNLINK, errno, FLN, PN_(dst, 1));
	}
	
#ifdef HAVE_READLINK
	if( readlink(src, link_buf, PATH_LENGTH) < 0 ) {
		/* Not a symbolic link */
		if( errno == EINVAL ) {
			return 1;
		/* Error occurred */
		} else {
			error_out(ERROR_ERROR_RDLNK, errno, FLN, PN_(src, 1));
			return -1;
		}
	}
#else
#  warning readlink() is not supported. 
	strncpy(link_buf, src, PATH_LENGTH);
#endif
	
	verbose(VERBOSE_LOW, "File is a symbolic link:\nfile=\"%s\"", src);
	verbose(VERBOSE_MID, "Read symbolic link:\nsymlink=\"%s\"\nread=\"%s\"", src, link_buf);
	
#ifdef HAVE_SYMLINK
	/* Try to creat symlink at dst with contents of symlink file at src.
	 * New symlink may turn out to be a dangling link. 
	 */
	if( symlink(link_buf, dst) < 0 ) {
		error_out(ERROR_ERROR_CSYMLNK, errno, FLN, PN_(dst, 1));
		return -1;
	}
#else
#  warning symlink() is not supported.
#endif
	
	/* Preserve mtime and atime */
	if( equate_am_time(src, dst, TRUE) < 0 ) {
		error_out(ERROR_ERROR_SLEQMTIME, errno, FLN, src, PN_(dst, 1));
		return -1;
	}
	
	return 0;	
}
