/*
 *      file.h - General file handling routines
 *		Part of the usbsync utility 
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifndef FILE_H
#define FILE_H 1

#include <sys/stat.h>

#include "common.h"


/**
 * General structure for a USBS_FILE, that is two files, one on the usb medium 
 * and one on the hd. These then will be synchronized with each other.
 */
typedef struct {
	
	/* File names on usb and host */
	char *usb_fname;
	char *host_fname;
	
	/* uid file belongs to on host */
	uid_t uid;
	/* gid */
	gid_t gid;
	
	/* Mode bits of the more recent file */
	mode_t mod;
	
	/* Last modified times */
	time_t usb_time;
	time_t host_time;
	
	/* Only used for directories w/o content */
	//BOOL is_empty_dir;
	
	/* File type determines kind of file */
	enum usbs_filetype {
		NORMAL_FILE = 0,
		NORMAL_DIR,
		EMPTY_DIR,
		/* Every directory, after it has been created as a
		 * normal or empty dir must get updated (mtime set up-to-date).
		 * Therefore, every dir will become an "UPDATE_DIR" eventually.
		 * They will be placed on a separate stack and synchronized in the end.
		 * See sync.c (directory stack).
		 */
		UPDATE_DIR, 
	} file_type;
	
} USBS_FILE;

extern char *file_type_names[];

/* Prototypes */

extern USBS_FILE* file_new();
extern void file_destroy(USBS_FILE **file);

extern FILE *open_file_stream(const char *file, const char *mode, BOOL exit_on_failure);
extern int open_file_descriptor(const char *file, int flags, mode_t, BOOL exit_on_failure);
extern char *readline_stream(FILE *s);
extern int cpy_file(const char *src, const char *dst, int uid, int gid, BOOL exit_on_failure);
extern int cpy_symlink(const char *src, const char *dst, int uid, int gid, BOOL aggressive);
extern void get_directory(char *dir, const char *path);
extern void concat_dirs(char *end_dir, const char *dir1, const char *dir2);
extern void* usbs_fgets(void *buf, int size, FILE *fstream, BOOL exit_on_failure);
extern BOOL is_symlink(const char *path, BOOL silent);
extern BOOL is_dangling_symlink(const char *path);
extern BOOL chk_existence(const char *path);
extern int true_filesize(const char *path);
extern void symlink_info(char *buf, const char *descr, const char *path);
extern int usbs_chown(enum expand_type direction, const char *path, uid_t uid, gid_t gid);
extern int mode_equate(enum expand_type direction, const char *file2, const char *file1);
extern int equate_am_time(const char *src, const char *dst, BOOL only_symlink);
extern BOOL is_directory(USBS_FILE *f);
extern BOOL is_dir_ftype(enum usbs_filetype ftype);
extern int usbs_stat(char *f, struct stat *s);

#endif /* FILE_H */
