/*
 *      parser.c - usbsync file parsing and directory scanning core functions
 *      Part of the usbsync utility 
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#endif

#include "parser.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h> 
#include <sys/stat.h> 
#include <unistd.h>
#include <fcntl.h>
#include <libgen.h>
#include <unistd.h>
#include <assert.h>

/* Dirent portability */
#ifdef HAVE_CONFIG_H
#  ifdef HAVE_DIRENT_H
#    include <dirent.h>
#    define NAMLEN(dirent) strlen ((dirent)->d_name)
#  else
#    define dirent direct
#    define NAMLEN(dirent) ((dirent)->d_namlen)
#    if HAVE_SYS_NDIR_H
#      include <sys/ndir.h>
#    endif
#    if HAVE_SYS_DIR_H
#      include <sys/dir.h>
#    endif
#    if HAVE_NDIR_H
#      include <ndir.h>
#    endif
#  endif
#else
#  include <dirent.h>
#endif

#include "common.h"
#include "file.h"
#include "sync.h"
#include "debug/debug.h"
#include "stack/stack.h"
#include "error/usbs_errors.h"
#include "options/options.h"
#include "path.h"
						  
/*
 * Syncdata options (and corresponding .usbsync tokens) for a synchronization entry
 */
struct syncdata_options sync_opts[] = {
	{"tape_archive", TARBALL_ARCHIVE},      /* Synchronization entry is tar'd */
	{"gzip_compression", GZIP_COMPRESSION}, /* Apply gzip compression t0 tar archive*/
};


/**
 * Pushes, according to to enum expand_type direction being set,
 * a new USBS_FILE onto STACK s passed.
 * NOTE: All the char arrays should have been obtained thru a dynamic
 * memory allocation as references to them are stored in the new file
 * object.
 */
static void push_new_file(
	STACK *s,
	enum expand_type direction,
	enum usbs_filetype ftype,
	char *source,
	char *destin,
	int src_mtime,
	int dest_mtime,
	USER_INFO uinfo) 
{
	USBS_FILE *file;
	file = file_new();
	
	++files_check;
	
	/* is src=usb and dest=hd or the other way? */
	if( direction == USB_TO_HD ) {
		verbose(VERBOSE_MID, "Pushing new file:\nusb:\"%s\"\nhd: \"%s\"", 
		basename(source), basename(destin));
		file->usb_fname = source;
		file->host_fname = destin;	
		file->usb_time = src_mtime;
		/* > 0 means that dest file exists */
		if( dest_mtime > 0  )
			file->host_time = dest_mtime;
	} 
	else if( direction == HD_TO_USB ) {
		verbose(VERBOSE_MID, "Pushing new file:\nhd:\"%s\"\nusb:\"%s\"", 
		basename(source), basename(destin));
		file->host_fname = source;
		file->usb_fname = destin;
		file->host_time = src_mtime;
		if( dest_mtime > 0  ) 
			file->usb_time = dest_mtime;	
	} 
	 
	file->uid = uinfo.uid;
	file->gid = uinfo.gid;
	file->file_type = ftype;
	
	stack_push(s, file);
}		


/**
 * Is passed a (top-level) directory. This then is recursively traversed 
 * (post order) and this routine is called to expand the stack passed.
 * This means that all files encountered on the way are pushed onto
 * the file stack given. 
 * push_duplicates indacates that, if set, files that
 * might exist within both basedirectores+current_dir are pushed
 * onto the stack. If set false, they will not be pushed.
 */
static void expand_stack(
	int dir_rec_depth,
	enum expand_type action,
	STACK *s,
	USER_INFO uinfo,
	const char source_base_dir[],
	int src_bas_len,
	const char destin_base_dir[],
	int dst_bas_len, 
	const char current_dir[],
	BOOL push_duplicates)
{

/* Output a warning at this dir recursion; if `-e` not set */
#define LARGE_DIR_REC 30
	
	STACK *directs;
	DIR *dir;
	BOOL empty_dir;
	
	/* Relevant for LARGE_DIR_REC boundary */
	static enum {NORMAL = 0, SKIP, CONTINUE} recur_action; 
	
	struct dirent *dfile;
	struct stat fstat, fstat2;
	char current_src_dir[PATH_LENGTH], current_dest_dir[PATH_LENGTH], 
		 current_src_file[PATH_LENGTH], current_dest_file[PATH_LENGTH], 
		 new_current_dir[PATH_LENGTH]; 
	char *str_src, *str_dest;
	size_t current_dir_len;
		 
	verbose(VERBOSE_MID, "Dir-recursion: entering level: %d", dir_rec_depth);	
	
	/* New directory stack */
	directs = stack_new(), directs->stack_type = NORMAL_STACK;
	
	empty_dir = TRUE;
	*new_current_dir = *current_src_dir = *current_dest_dir = 0;   

	/* Keep track of path length; used to prevent overflowing path buffers */
	/* Add up path lengths */
	src_bas_len += strlen(current_dir);
	dst_bas_len += strlen(current_dir);  /* FIXME: needed ? */
	
	current_dir_len = strlen(current_dir);
		
	/* Prevent too long path names from overflowing path buffers */
	if( !check_path_length("Directory", source_base_dir, current_dir_len) ||
		!check_path_length("Directory", destin_base_dir, current_dir_len) ) return;

	concat_dirs(current_src_dir, source_base_dir, current_dir);
	concat_dirs(current_dest_dir, destin_base_dir, current_dir);
	
	/*
	FIXME:
	verbose(VERBOSE_MID, "current_src_dir=\"%s\"", PN_(current_src_dir, 0));
	verbose(VERBOSE_MID, "current_dst_dir=\"%s\"", PN_(current_dest_dir, 0));
	*/
	
	++dirs_check;

	/* Current_src_dir always exists whereas current_dest_dir is hypothetical
	 * depending on the overall contents of the source_bas_dir
	 */
	
	if( !(dir = opendir(current_src_dir)) ) {
		error_out(ERROR_ERROR_DIROPN, errno, FLN, PN_(current_src_dir, 0));
		return;
	}
	
	/* Go through directory contents */
	while ( (dfile = readdir(dir)) != NULL ) {
		
		/* Skip "." and ".." */
		if( !strcmp(dfile->d_name, ".") || !strcmp(dfile->d_name, "..") ) 
			continue;
			
		if( !check_path_length("In-directory", current_src_dir, strlen(dfile->d_name)) )
			continue;
		
		/* Assemble path of file or directory in this dir */
		concat_dirs(current_src_file, current_src_dir, dfile->d_name);
		
		/* Stat information */
		if( lstat(current_src_file, &fstat) < 0 ) {
			error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(current_src_file, 1));	
			continue;
		}
		
		if( S_ISDIR(fstat.st_mode) ) {
			empty_dir = FALSE;
			/* If this directory is a symlink and the `-l` option is set the directory will not
	   		   be expanded.
	 		 */
			if( is_symlink(current_src_file, 1) && _option_check(_OPTION_CPY_SYMLINKS) ) { 
				verbose(VERBOSE_LOW, "Directory is a symbolic link; will no be expanded:\n\"%s\"", 
				PN_(current_src_file, 0));
				/* Jump to adding directory as a normal file; 
				   it will later on be detected just to be
				   a symbolic link and will be copied accordingly.
				 */
				goto symlink_dir;
			}
			/* Push directory onto directory stack for subsequent recursive calls following*/
			verbose(VERBOSE_HIGH, "Pushing directory:\n\"%s\"", dfile->d_name);
			stack_push(directs, (void*)cpy_string(dfile->d_name));
		} else {
			symlink_dir:
			/* Fill state structure again, now with lstat */
			if( lstat(current_src_file, &fstat) < 0 ) {
				error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(current_src_file, 1));	
			}
			
			/* File is not an empty directory */
			empty_dir = FALSE;
			
			/* Destination file or directory path */
			if( !check_path_length("In-directory", current_dest_dir, strlen(dfile->d_name)) )
				continue;
			concat_dirs(current_dest_file, current_dest_dir, dfile->d_name);
			
			fstat2.st_mtime = 0;
			/* Destination file already exists or is a symbolic link */
			if( access(current_dest_file, 0) == 0 || is_symlink(current_dest_file, 1) ) {
				verbose(VERBOSE_MID, "File or directory already exists:\n\"%s\"", 
				PN_(current_dest_file, 0));
				/* Pushing duplicates not allowed */
				if( !push_duplicates ) {
					verbose(VERBOSE_HIGH, "Skipping duplicate files:\n\"%s\"\n\"%s\"", 
					PN_(current_src_file, 1), PN_(current_dest_file, 1));
					continue;	
				}
				if( lstat(current_dest_file, &fstat2) < 0 ) {
					error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(current_dest_file, 1));	
					continue;	
				}
			} 
			verbose(VERBOSE_HIGH, "Pushing new soure file:\n\"%s\"", PN_(current_src_file, 1));
			
			str_dest = cpy_string(current_dest_file);
			str_src = cpy_string(current_src_file);
			
			/* Push new file according to src and dest */
			push_new_file(s, action, FALSE, str_src, str_dest, 
			fstat.st_mtime, fstat2.st_mtime, uinfo);
			
		}
	}
	
	if( verbosity_check(VERBOSE_HIGH) ) {
		msg("Directory stack at recursive level %d:\v", dir_rec_depth);
		dump_stack(directs, STR);
	}
	
	++dir_rec_depth;
	
	/* Check directory recursion level first before continuing (-d option) */
	if( dir_recursion > -1 && dir_rec_depth >= dir_recursion ) {
			verbose(VERBOSE_LOW, "Break at recursion level %d:\n\"%s\"",
			dir_rec_depth,  current_src_dir);
			goto skip_next_recursion;
	}
	
	/* Check if next directory recursion level would not exeed default max level depth */
	if( dir_rec_depth >= LARGE_DIR_REC ) {
		if( recur_action == NORMAL ) {
			error_out(ERROR_WARNING_DRCBN, 0, FLN, dir_rec_depth);
			if( prompt_question("Do you want to continue beyond this boundary?", "YN", &msg) == 'N' ) {
				msg("Breaking at recursive level: %d\v", dir_rec_depth);
				recur_action = SKIP;
				goto skip_next_recursion;
			} else recur_action = CONTINUE;
		} else if( recur_action == SKIP )
			goto skip_next_recursion;
	}
	
	/* Now traverse sudirectories in directory stack 
	 * and call expand_stack recursively.
	 */
	for( ; directs->count > 0; stack_pop(directs, 1) ) {
		
		/* Assemble new directory path for next recursive call */
		if( !check_path_length("Directory", current_dir, strlen((char *)directs->top->data)) )
			continue;
		concat_dirs(new_current_dir, current_dir, (char *)directs->top->data);
		
		verbose(VERBOSE_MID, "Expanding file stack; recursive directory search:"
		"\nnew_current_dir=\"%s\"", PN_(new_current_dir, 0));
		expand_stack(dir_rec_depth, action, s, uinfo, source_base_dir, src_bas_len, destin_base_dir,
		dst_bas_len, new_current_dir, push_duplicates);
	}
	
	/* We get jumped here if we skip the next recursive level due to `-d` being set */
	skip_next_recursion:

	/* Prepare for pushing directory itself */		
	str_src = cpy_string(current_src_dir);
	str_dest = cpy_string(current_dest_dir);
	
	/* Retrieve information of this source directory */
	if( lstat(str_src, &fstat) < 0 ) {
		error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(current_dest_file, 0));
		free(str_src);
		free(str_dest);
		return;
	} 
	
	/* Look if destination directory exists */ 
	if( chk_existence(str_dest)) {
		/* Retrieve information for destination directory */
		if( lstat(str_dest, &fstat2) < 0 ) {
			error_out(ERROR_ERROR_LSTAT, errno, FLN, PN_(current_dest_file, 0));
			free(str_src);
			free(str_dest);
			return;
		}
	}
	/* Dest dir does not exist */
	else fstat2.st_mtime = 0;

	/* If other dir does not exist or "push duplicates" is allowed we push directory */
	if( fstat2.st_mtime == 0 || push_duplicates ) {
		
		/* Empty dirs and normal dirs get pushed as a particual kind of file now */
		push_new_file(s, action, empty_dir ? EMPTY_DIR : NORMAL_DIR, str_src, str_dest, 
		fstat.st_mtime, fstat2.st_mtime, uinfo);
		
		verbose(VERBOSE_HIGH, "Pushing directory as file:\n\"%s\"\n\"%s\"", 
		PN_(str_src, 0), PN_(str_dest, 0)); 	
		
	} else {
		/* Cleanup if not pushed */
		free(str_src);
		free(str_dest);
	}
	
	/* Cleanups */
	stack_destroy(&directs);
	free(dir);

}

/**
 * Parses and expession like xxxxxx:"a", "b", "c", ... 
 * The values parsed, that is the strings, "a", "b", "c"
 * are pushed onto a stack. A reference to this stack is returned.
 * The character sep, usually a double quotation mark, is taken as 
 * the delimeters of each token to be parsed.
 * NOTE: *string passed must be \0-terminated!
 */
static STACK* psfix_parse(const char *string, char sep) {
	STACK *fixes = stack_new();
	char *beg, *end, str[512];
	int qcount; /* Quotation marks count */
	
	strncpy(str, string, sizeof str);
	beg = end = str;
	
	while(*end) {
		qcount = 0;
		
		while(*++end != ',' && *end != '\0')
			if(*end == sep) ++qcount;
			
		/* Syntax error: Insufficient quotation marks counted */
		if(qcount < 2) {
			stack_destroy(&fixes);
			return NULL;
		}
		
		beg = strchr(beg, sep);
		end[-1] = 0;
		
		if( *++beg ) stack_push(fixes, (void*)cpy_string(beg));
		
		beg = end;
	}
	
	return fixes;
}

/**
 * Returns a reference to a USBS_SYNCTDATA structure which will contain a STACK of USBS_FILE files.
 * that is the files that have been scanned and that are to be synchronized. The structure returned
 * will also contain as tack of allowed file/dir prefices suffices.
 * The pointer returned is nerver NULL, rathter the stacks within can be NULL.
 */
STACK* parse_usbsync(char *dev_mountpoint) {
/* Macro to advance string pointer to character and to obtain a string as parsing result */
#define PARSE(STR, P, C) while(*++P != C && *P ); *P = 0; STR = P + 1
		
	char usbsync[PATH_LENGTH], tmp_buf[BUF_MAX_LEN];
	char usb_file[PATH_LENGTH], path[PATH_LENGTH], start_info[BUF_MAX_LEN];
	char *line, *last_usb_file_line, *host_file;
	char *user, *host, *sp, *eof;
	struct stat fstat;
	FILE *sfile;
	int line_cnt, last_usb_file_line_cnt = -1, tmp;
	BOOL require_next, host_found, dir_error;
	mode_t tmpmode;
	BOOL file_available = FALSE;
	uid_t euid;
	USER_INFO uinfo;
	
	/* New stack of syncdata structures */
	STACK *syncdata_stack = stack_new();
	syncdata_stack->stack_type = SYNCDATA_STACK;
	
	/* syncdata structure */
	USBS_SYNCDATA *sync_data = NULL;
	
	/* usbs file template */
	USBS_FILE *f = NULL;
	
	/* Check if mountpoint path is within length boundaries */
	if( !check_path_length("Mount point", dev_mountpoint, strlen(usbsyncfile_name)) ) {
		stack_destroy(&syncdata_stack);
		goto finish;
	}
	
	/* Obtain .usbsync file path */
	concat_dirs(usbsync, dev_mountpoint, usbsyncfile_name);
	
	/* Initialize parsing booleans and variables */
	require_next = FALSE;
	host_found = TRUE;
	last_usb_file_line = NULL;
	tmpmode = (mode_t)0;
	*start_info = 0;

	/* Open usbsync file */
	if( (sfile = open_file_stream(usbsync, "r", FALSE)) != NULL ) { 
		
		usr_host_info(start_info);
		msg("Synchronization on \"%s\" started\n%s\v", PN_(dev_mountpoint, 0), start_info);
		draw_separator('=');
		__out();
		sprintf(tmp_buf, "(Max directory recursion level: %d) ", dir_recursion);
		msg("Scanning files and directories %s... \v", dir_recursion > -1 ? tmp_buf : EY); 
		
		/* Now go through .usbsync file line by line */
		eof = (char*)TRUE;
		line_cnt = 0;
		
		while( eof ) {
			
			/* Time measuring, always per line */
			time_needed += __stop_timer();
			__start_timer();
			
			/* Read next line */
			line = eof = readline_stream(sfile);
	
			/* EOF has been reached */
			if( !eof ) {
				verbose(VERBOSE_HIGH, "Reached EOF: \"%s\"", usbsync);
				/* There still need some checks to be performed regarding last usb file, 
				 * so jump here.
				 */
				goto eof;
			}
			
			/* Increment line count */
			++line_cnt;
			
			/* Skip leading white spaces */
			while( *line == ' ' || *line == '\t' ) line++;
			
			/* Skip comments and empty lines */
			if( *line == '#' || *line == '\n' ) continue;
			
			/* Cut off \n at the end of lines as returned by fgets */
			line[strlen(line)-1] = '\0';
			
			/* New file header denoting usb file */
			if( !strchr(line, ':') && *line != '+' ) {				
				
				/* When eof reached, perform last checks */
				eof:
				/* Check if any association was set at all for file on usb */
				if( require_next ) {
					error_out(ERROR_ERROR_FASSOC, 0, NULL, 0, usbsyncfile_name, 
					last_usb_file_line_cnt, PN_(last_usb_file_line, 0)); 
					DESTROY(file_destroy, f);
					DESTROY(syncdata_destroy, sync_data);
				}
				
				/* Check if a valid host was found to synchronize with */
				if( !host_found && !require_next) {	
					error_out(ERROR_ERROR_NOHOST, 0, NULL, 0, 
					usbsyncfile_name, last_usb_file_line_cnt, PN_(last_usb_file_line, 0));
					DESTROY(file_destroy, f);
					DESTROY(syncdata_destroy, sync_data);
				}
				
				
				/* If we are in a new section and syncdata != NULL we need to:
				   set it to NULL and push it onto the syncdata stack */
				if( sync_data != NULL ) {
					
					/* Nothing on file stack */
					if(sync_data->sync_files->count < 1) {
						stack_destroy(&sync_data->sync_files);
					}
	
					/* Add suffix for update file to ignore stack */
					if( !sync_data->suffix_ign ) sync_data->suffix_ign = stack_new();
					stack_push(sync_data->suffix_ign, cpy_string(BACKUPFILE_SUFFIX));
					
					/* Push syncdata itself */
					stack_push(syncdata_stack, (void*)sync_data);
					sync_data = NULL;
				}
				
				/* If EOF has been reached break after these checks, and destroy file obj */
				if( !eof ) { 
					/* File is available here if it has not been pushed nor destroyed already */
					if( file_available ) file_destroy(&f);
					DESTROY(syncdata_destroy, sync_data);
					verbose(VERBOSE_HIGH, "EOF cleanups have been performed");
					break;
				}
				
				/* Remember last usb file line */
				last_usb_file_line_cnt = line_cnt;
				if( last_usb_file_line ) free(last_usb_file_line);
				last_usb_file_line = cpy_string(line);
			
				/* Validate syntax of path, that is, a relative path */
				if( !validate_usb_path(line) ) {
					error_out(ERROR_ERROR_SYNSEM, 0, NULL, 0, usbsyncfile_name, line_cnt);
					continue;
				}

				/* Check path length */
				if( !check_path_length("Storage volume", dev_mountpoint, strlen(line)) )
					continue;
			
				concat_dirs(usb_file, dev_mountpoint, line);
				
				/* New usbs file */
				f = file_new();
				file_available = TRUE;
				
				/* New syncdata structure containing file stack and */
				sync_data = syncdata_new();
				sync_data->sync_files = stack_new();
				sync_data->sync_files->stack_type = FILE_STACK;
				sync_data->usb = cpy_string(usb_file);
				
				if( stat(usb_file, &fstat) < 0 ) {
					verbose(VERBOSE_LOW, " ** %s:%d: Non-existing file:\n\"%s\"", 
					usbsyncfile_name, line_cnt, PN_(usb_file, 1));
				} else {
					verbose(VERBOSE_MID, " ** %s:%d: File found:\n\"%s\"", 
					usbsyncfile_name, line_cnt, usb_file);
					
					/* Copy usb file modification time */
					f->usb_time = fstat.st_mtime;
					/* Save file mode */
					tmpmode = fstat.st_mode;
				}
				
				/* Copy usb file name to USBSYNC_FILE object */
				f->usb_fname = cpy_string(usb_file);
				require_next = TRUE;
				host_found = FALSE;
				
			/* Line denoting a corresponding file or directory contains '@' and ':' */
			} else if( strchr(line, '@') && strchr(line, ':') ) {
				
					/* If no file is available this section is skipped */
					if( !file_available ) continue;
					
					/* Next found for usb file; none required anymore */
					require_next = FALSE;

					/* Parse user, host, and host_file from line */
					user = host = host_file = NULL;
					sp = user = line;
					PARSE(host, sp, '@');
					PARSE(host_file, sp, ':');
					
					/* A host was found */
					host_found = TRUE;
						
					verbose(VERBOSE_MID, "Parsing results:\n"
					"file=\"%s\":\nuser=\"%s\", host=\"%s\"\nhost_file=\"%s\"", 
					 FN_(usb_file), user, host, PN_(host_file, 1));
					 
					/* Check parsing results */
					if( !*user || 
					    !*host || 
					    !*host_file ||
					    !validate_host_path(host_file)
					) {
						/* Error occurred with parsing stuff */						
						error_out(ERROR_ERROR_SYNSEM, 0, NULL, 0, usbsyncfile_name, line_cnt);
						
					/* Compare this host's name with host name for file or directory */
					} else if( (strncmp(this_hostname, host, strlen(host)) == 0) ) {
						
						/* Get user information */
						uinfo = getuserinfo(user);
						
						/* Output warning if executing user is not root and is 
						   different from user parsed */
						euid = geteuid();
						if( euid != 0 && euid != uinfo.uid ) {
							error_out(ERROR_WARNING_UPRIV, 0, FLN, 
							euid, uinfo.uid, this_username);
						}
						
						/* Expand host_file if relative path with '~' character is given to user's
						 * home directory 
						 */
						if( !(host_file = expand_tilde_path(path, host_file, uinfo.uid)) )
							continue;
							
						sync_data->host = cpy_string(host_file);
							
						/* Check if file exists on host */
						if( stat(host_file, &fstat) < 0 ) {
							
							/* Corresponding file on host not yet present */
							verbose(VERBOSE_HIGH, 
							"%s:%d:\n\"%s\"\nAssociation on \"%s\" not found",
							usbsyncfile_name, line_cnt, PN_(host_file, 1), host);
								
						} else {
							
							/* Everything OK */
							verbose(VERBOSE_MID, " ** %s:%d: File found:\n\"%s\"", 
							usbsyncfile_name, line_cnt, PN_(host_file, 1));
							verbose(VERBOSE_LOW, "File or directory association made:\nusb:\"%s\"\n"
							"%s:\"%s\"", basename(usb_file), host, basename(host_file));
							
							/* Copy host file modification time */
							f->host_time = fstat.st_mtime;
						} 
						
						verbose(VERBOSE_LOW, "Corresponding file or directory pair found:"
						"\nusb=\"%s\"\n%s=\"%s\"", 
						PN_(f->usb_fname, 0), host, PN_(host_file, 0));
						
						/* Copy host file information */
						f->host_fname = cpy_string( host_file );
						
						/* Now set uid and gid for user parsed */
						f->uid = uinfo.uid;
						f->gid = uinfo.gid;
						
						/* Look if new file is to be pushed as it migh be
						   a real file and no dir */
						if( f != NULL ) {
							
							/* Presume that it is not a directory */
							dir_error = FALSE;
							
							/* If usb file exists and is a directory*/
							if( f->usb_time > 0 && S_ISDIR(tmpmode) ) {
								/* If host file exists but is no directory we have an 
								   association error */
								if( f->host_time > 0 && !S_ISDIR(fstat.st_mode) ) {
								   error_out(ERROR_ERROR_DIRFILE, 0, FLN,
								   usbsyncfile_name, line_cnt, 
								   PN_(f->usb_fname, 0), PN_(f->host_fname, 0));
								   dir_error = TRUE;		
								}
							}
							
							/* If host file exists and is a directory*/
							if( f->host_time > 0 && S_ISDIR(fstat.st_mode) ) {
								/* If usb file exists but is no directory; error */
								if( f->usb_time > 0 && !S_ISDIR(tmpmode) ) {
								   error_out(ERROR_ERROR_DIRFILE, 0, FLN,
								   usbsyncfile_name, line_cnt, 
								   PN_(f->host_fname, 0), PN_(f->usb_fname, 0));
								   dir_error = TRUE;	
								}
							}
							
							/* Both "files" are dirs or prospective dirs */
							if( !dir_error ) {  
							
								if( S_ISDIR(fstat.st_mode) || S_ISDIR(tmpmode) ) {
									
									verbose(VERBOSE_MID, 
									"Files are directories:\n\"%s\"\n\"%s\"", 
									basename(host_file),	
									basename(usb_file));
									
									/* If dir exists on usb, expand dir 
									   to all files that exists within in
									   (including subdirs).
									   All those files will be pushed
									   ontop of the file stack.
									 */
									if( f->usb_time > 0 ) {
										expand_stack(0, USB_TO_HD, sync_data->sync_files,
										uinfo, f->usb_fname, strlen(f->usb_fname),
										f->host_fname, strlen(f->host_fname), 
										"", FALSE);
									}
									
									/* If dir exist on host do also traverse
									 * and check
									 */
									if( f->host_time > 0 ) {
										expand_stack(0, HD_TO_USB, sync_data->sync_files, 
										uinfo, f->host_fname, strlen(f->host_fname),
										f->usb_fname,  strlen(f->usb_fname), 
										"", TRUE);
									} 
									
									/* Both files/dirs on host and usb do not exist; error */
									if ( f->usb_time < 1 && f->host_time < 1 ) {
										error_out(ERROR_ERROR_EXIST, 0, FLN, 
								    	usbsyncfile_name, line_cnt, 
								    	PN_(usb_file, 0), PN_(host_file, 0));
									}
									
									/* 
									   NOTE: Be aware, subdirectories of target files 
									         might not exist.
									         Take care of this in sync process!
									 */
									
									/* We're done with this file */
									file_destroy(&f);	
									file_available = FALSE;
									
								/* Both on usb and hd are true files (no directories) */
								} else {
									
									/* Pushing files onto stack */
									if( f->usb_time > 0 || f->host_time > 0 ) {
										stack_push(sync_data->sync_files, (void*)f);
										file_available = FALSE;
										++files_check;
									} else {
								    	error_out(ERROR_ERROR_EXIST, 0, NULL, 0, 
								    	usbsyncfile_name, line_cnt, 
								    	PN_(usb_file, 0), PN_(host_file, 0));
								    	file_destroy(&f);
								    	file_available = FALSE;
									}	
								}
							} else {
								/* We're done with this file */
								file_destroy(&f);
								file_available = FALSE;
							}	
						}
					}
			/* Another line is a ignore postfix/suffix specifier for directories */	
			} else if( strchr(line, ':') && strchr(line, '\"') ) {
				
				 if( !host_found ) continue;
				 
				 /* Prefix ignore string found */
				 if( strncmp(line, IGNORE_PREFIX_STRING, strlen(IGNORE_PREFIX_STRING)) == 0 ) {
				 	if( (sync_data->prefix_ign = psfix_parse(line, '\"')) == NULL ) {
						error_out(ERROR_ERROR_SYNLIN, 0, NULL, 0, 
						usbsyncfile_name, line_cnt, line);	
					}
				 	
				 } 
				 /* Suffix ignore string found */
				 else if( strncmp(line, IGNORE_SUFFIX_STRING, strlen(IGNORE_SUFFIX_STRING)) == 0 ){
				 	if( (sync_data->suffix_ign = psfix_parse(line, '\"')) == NULL ) {
				 		error_out(ERROR_ERROR_SYNLIN, 0, NULL, 0, 
						usbsyncfile_name, line_cnt, line);
					}
				/* Error: invalid identifier */
				} else {
					error_out(ERROR_ERROR_SEMLIN, 0, NULL, 0, usbsyncfile_name, line_cnt, line);
				}
				
			} else if( *line == '+' ) {
				
				BOOL option_set = FALSE;
				++line;
				
				/* Check for other options */
				for( tmp = 0; tmp < SIZE_OF(sync_opts); tmp++ ) {
					if( strncmp(line, sync_opts[tmp].option_token, 
						strlen(sync_opts[tmp].option_token)) == 0 ) {
						sync_data->flags |= sync_opts[tmp].flag;
						option_set = TRUE;
					}	
				}
				
				if( !option_set ) 
					error_out(ERROR_ERROR_IOPT, 0, NULL, 0, usbsyncfile_name, line_cnt, line);
						
			} else {
					error_out(ERROR_ERROR_MALFRM, 0, NULL, 0, usbsyncfile_name, line_cnt, line);	
			}
			
		}
		fclose(sfile);	
	} else {
		/* Cannot open .usbsync file */
		error_out(ERROR_FATAL_PARSING, 0, FLN, usbsync);
	}
	
	/* Free last line buffer */
	if( last_usb_file_line ) free(last_usb_file_line);
	
	/* Free static memory that held lines */
	readline_stream(NULL);
	
	finish:
	return syncdata_stack;
	
} 
