/*
 *      path.c - Working with path, file and directory names
 *		Part of the usbsync utility 
 * 		
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#endif

#include "path.h"

#include <stdio.h>
#include <stdlib.h>
#include <libgen.h>

#include "common.h"
#include "debug/debug.h"
#include "error/usbs_errors.h"

/**
 * Lengths for the maximum length of a filename and a path to be 
 * displayed by usbsync. They are generally depending on th
 * termcap value COLS, if retrievable. Otherwise default values will
 * be used. They are set by __init_path_display().
 * This is done in order to avoid ugly looking things on the terminal.
 */
int MAX_FNAME_LEN;
int MAX_PATHN_LEN;

/**
 * Is passed tilde_path (a null terminated string of the form ~/<path>)
 * where '~' (tilde) stands for the user specified by uid.
 * Returns pointer to tilde_path if tilde_path is not a valid path 
 * preceeded by a tilde character. 
 * Returns pointer to buffer abs_path, which needs to be specified on 
 * the call.
 * Returns NULL if an error occurred and tilde path could not be
 * expanded.
 */
char* expand_tilde_path(char *abs_path, char *tilde_path, uid_t uid) {
#define _HOMEDIR_CHAR '~'
	char home_dir[PATH_LENGTH];
	
	if( *tilde_path != _HOMEDIR_CHAR ) return tilde_path;
	
	*abs_path = 0;
	
	if( !get_user(HOME_DIR, home_dir, uid) ) {
		error_out(ERROR_ERROR_PATHEX, 0, FLN, (int)uid, tilde_path);
		return NULL;
	}
	
	++tilde_path;
	if( strlen(home_dir) + strlen(tilde_path) > PATH_LENGTH ) {
		error_out(ERROR_ERROR_PATHLEN, 0, FLN);
		return NULL;	
	}
	strcat(home_dir, tilde_path);
	strcpy(abs_path, home_dir);
	
	verbose(VERBOSE_MID, "Relative path successfully expanded\nrel=\"%s\"\nabs=\"%s\"",
	tilde_path, abs_path);
	
	return abs_path;
}


/**
 * Shortens a file name to be displayed
 * ===
 * This function is called for any file name string that is output by
 * this application. If the filename exceeds a certain length
 * (MAX_FNAME_LEN) it is shortened according to the following rules:
 * If a suffix after the '.' exists, it is preserved.
 * Then, all other characters from the beginning of the file name string
 * are shortened so that they fit the maximal length.
 * A tilde character ('~') is inserted to represent the characters not
 * shown. 
 * Example:  a_much_too_long_file_name.bla.txt becomes
 *           a_much_to~.txt
 */
char* FN_(const char *fname) {
#define FILE_SUFFIX_SEP '.'
#define TOO_LONG_SUBST  '~'
	
	char suffix[NAME_LENGTH];
	char filename[NAME_LENGTH];
	char *r = (char*)fname, *sf;
	int i;
	
	if( !*r || strlen(r) <= MAX_FNAME_LEN ) goto finish;
	
	r += strlen(fname)-1, sf = suffix;
	
	/* Preserve suffix */
	do { 
		*sf++ = *r;
		/* No suffix found */
		if( r <= fname ) {
			*suffix = 0; /* FIXME: Check suffix length to prevent buffer overflow */
			break;
		}
	} while( *r-- != FILE_SUFFIX_SEP );
	
	*sf = 0;
	
	str_reverse(suffix);
	
	/* Copy file until max valid length */
	for( i = 0; i < MAX_FNAME_LEN-strlen(suffix)-1  && *(fname+i); i++ )
		filename[i] = fname[i];
	
	/* Substitute overlaying characters */
	filename[i]   = TOO_LONG_SUBST; 
	filename[++i] = 0;
	
	/* Concatenate shortened file name and suffix (if there is one) */
	if( usbs_strcat(filename, suffix, sizeof filename) == NULL ) {
		error_out(ERROR_ERROR_FILELEN, 0, FLN);
	}
	
	r = new_str(filename);
	
	finish:
	return r;
}


/**
 * Concatenate multiple directories given in variable argument list with
 * delimiter given.
 * NOTE: Make sure everthing is correct with buf; no checks are performed herein.
 */
char* concat_m_dirs(char *buf, int n, ...) {
/* FIXME: Make this method saver to buffer overflows and 
 * use usbs_strcat(buf, delim, size_t len)
 */
	int i;
	va_list l;
	char *next_dir, delim[2] = {PATH_NAME_DELIM, 0};
	
	va_start(l, n);
	
	for( i = 0; i < n; i++ ) {
		next_dir = va_arg(l, char*);
		if( next_dir == NULL ) continue;
		strcat(buf, next_dir);
		strcat(buf, delim);
	}
	
	/* Cut of trailing '/' */
	last_chr(buf);
	
	va_end(l);
	
	return buf;
}

/**
 * Shorten a path which is of type path_type.
 * A path is shortened by taking out dirs in the middle and substituting
 * them with `...' if the whole path would otherwise exeed path_max_len.
 * If path referers to a file and is_file is set or if the file 
 * exists anyways, FN_() is used to also shorten this file name.
 * A rerefence to a string (using new_str()) is returned on success,
 * otherwise a reference to path itself is returned.
 */
char* PNN_(const char *path, enum path_type ptype, size_t path_max_len, BOOL is_file) {

	char p[PATH_LENGTH], pa[MAX_PATHN_LEN], dir[NAME_LENGTH], pathname[PATH_LENGTH];
	char *fname = (char*)0, *r = (char*)path, *last_dir;
	int i, len_dir, len_fil = 0, len_tot = 0;
	struct stat s = { .st_mode = (mode_t)0 };
	
	if( path_max_len < 1 ) path_max_len = MAX_PATHN_LEN;
	
	/* Path must be of form `/<dir1>/<dir2>/<dir3 or file>' for absolute,
	   `<dir1>/<dir2>/<dir3 or file>' for relative, or 
	   `~/<dir1>/<dir2>/<dir3 or file>' for home bath */
	
	if( !*path || strlen(path) < path_max_len || !validate_path_syntax(path, ptype) ) 
		goto finish;
	
	switch(ptype) {
		case ABSPATH:
		case HOMEPATH:
			if( strcnt(PATH_NAME_DELIM, (char*)path) < 3 )
				goto finish;
		break;
		case RELPATH:
			if( strcnt(PATH_NAME_DELIM, (char*)path) < 2 )
				goto finish;
		break;
	}
		
	/* Retrieve file information */
	if( !is_file && lstat(path, &s) );
	
	/* Copy absolute paths without leading '/' or*/
	switch(ptype) {
		case RELPATH:
			strncpy(p, r, PATH_LENGTH);
		break;
		case ABSPATH:
			strncpy(p, r+1, PATH_LENGTH);
		break;
		case HOMEPATH:
			strncpy(p, r+2, PATH_LENGTH);
		break;
	}
	
	bzero(pathname, sizeof pathname);
	
	/* If path referes to a file ... */
	if( S_ISREG(s.st_mode) || is_file) {
		r = basename(p);
		/* ... shorten filename */
		fname = FN_(r);
		/* ... cut off from path */
		p[ strlen(p) - strlen(r) - ( p[strlen(p)] == PATH_NAME_DELIM ? 0 : 1 ) ] = 0;
		len_fil = strlen(fname);
	}
	
	/* Extract last directory in path */
	last_dir = basename(p);
	len_dir = strlen(last_dir);
	
	/* Cut off last dir */
	p[ strlen(p) - strlen(last_dir) - 1 ] = 0;
	
	*pathname = *pa = 0;
	
	/* Parse single directory tokens */
	for( i = 1; *r; i++ ) {
		r = token_at_r(dir, PATH_NAME_DELIM, p, i);
		len_tot += strlen(r) + len_dir;
		/* If new path is too long cut off here; note the two extra '/' */
		if( len_tot + len_fil + 2 >= path_max_len ) {
			strcat(pa, DIRS_SHORTENED);
			break;
		}
		strcat(pa, r);
		chrcat(pa, PATH_NAME_DELIM);
	}
	
	if( strlen(pa) + strlen(last_dir) + len_fil >= PATH_LENGTH ) {
		error_out(ERROR_ERROR_PATHLEN, 0, FLN);
		goto finish;
	}
	
	/* Prepend proper path preliminary */
	if( ptype == ABSPATH ) {
		chrcat(pathname, PATH_NAME_DELIM);
	}
	else if( ptype == HOMEPATH ) {
		chrcat(pathname, PATH_HOME_INIDC);
		chrcat(pathname, PATH_NAME_DELIM);	
	}
	
	/* Put path back together */
	concat_m_dirs(pathname, 3, pa, last_dir, fname);
	
	/* Obtain a string stored on heap */
	r = pathname;
	
	finish:
	return new_str(r);
}

/**
 * Shorten an absolute pathname to fit the maximum length of
 * MAX_PATHN_LEN (should be adopted to be pending on LINE).
 * Override check if path referes to file by passing is_file as TRUE.
 */
char* PN_(const char *path, BOOL is_file) {
	return PNN_(path, *path == '/' ? ABSPATH : RELPATH, MAX_PATHN_LEN, is_file);
}

/* String name of path types for output */
const char *path_type_names[] = {"Host", "Volume"};

/**
 * The usual unix path resolution is restricted by usbsync. Path
 * resolution is important for specifying paths in the .usbsync file.
 * For further details see Unix/Linux path resolution manual page.
 */
BOOL validate_path_syntax(const char *path, enum path_type ptype) {
	char buf[NAME_LENGTH];
	int backrefn;

	/* Check path beginning */
	switch(ptype) {
		/* Absolut path */
		case ABSPATH:
			if( *path != PATH_NAME_DELIM  ) goto failure;
			++path;
		break;
		/* Relative path */
		case RELPATH:
			if( *path == PATH_NAME_DELIM  ) goto failure;
		break;
		/* Path starting with a tilde '~' refering to home dir*/
		case HOMEPATH:
			if( *path != PATH_HOME_INIDC ) goto failure;
			++path;
		break;
	}
	
	/* First token of absolute and relative path (the latter one for usbsync) may not be ".." */
	if( ( ptype == ABSPATH || ptype == RELPATH ) && 
		( strcmp(token_at_r(buf, PATH_NAME_DELIM , path, 1), "..") == 0 ) ) goto failure;
	
	/* Check if '..' goes along with number of directories */
	backrefn = token_count(path, "..", PATH_NAME_DELIM );
	if( gap_count(path, PATH_NAME_DELIM ) - backrefn < backrefn ) goto failure;
	
	/* All checks passed */
	return TRUE;
	
	/* Error handling */
	failure:
	return FALSE;
}

/**
 * Validates the syntax of a absolute or tilde path.
 * Returns false and outputs and error message upon failing.
 */
BOOL validate_host_path(const char *hpath) {
	if( validate_path_syntax(hpath, ABSPATH) || validate_path_syntax(hpath, HOMEPATH) )
		return TRUE;	
	error_out(ERROR_ERROR_PATHERR, 0, NULL, 0, path_type_names[0], PN_(hpath, 0));
	return FALSE;
}

/**
 * Validates the syntax of a relative path on the storage volume.
 * Returns false and outputs and error message upon failing.
 */
BOOL validate_usb_path(const char *upath) {
	if( validate_path_syntax(upath, RELPATH) )
		return TRUE;
	error_out(ERROR_ERROR_PATHERR, 0, NULL, 0, path_type_names[1], PN_(upath, 0));
	return FALSE;
}

/**
 * Validates a path length. tol gives a number of character that also have to
 * fit in meaning that overall, *path given must be exactly by that amount 
 * shorter than the usual path.
 * The path_name indicates what kind of path it is: this is important
 * for formatting the error message.
 * Returns FALSE upon failing and
 * outputs an error message.
 */
BOOL check_path_length(const char *path_name, const char *path, size_t tol) {
	if( strlen(path) >= PATH_LENGTH - tol ) {
		error_out(ERROR_ERROR_PATH, 0, FLN, path_name, path);
		return FALSE;
	}
	return TRUE;
}

void __init_path_display() {
	/* Relative factor to COLS, how long should a fname and pathname max be */
#define REL_PATH_FACT 0.5
#define REL_NAME_FACT 0.2
	
	/* Make sure terminfo is initialized */
	if( __init_termcap() < 0 ) {
		MAX_FNAME_LEN = MAX_FNAME_LEN_L;
		MAX_PATHN_LEN = MAX_PATHN_LEN_L;
	} else {
		MAX_FNAME_LEN = REL_NAME_FACT * cols;
		MAX_PATHN_LEN = REL_PATH_FACT * cols;	
	}	
	
#ifdef DEBUG
	verbose(VERBOSE_HIGH, 
	"Max displayed path name length set to: %d\n"\
	"Max displayed file name length set to: %d"
	, MAX_PATHN_LEN+1, MAX_FNAME_LEN+1);
#endif
}

void __deinit_path_display() {
	/* Free all stored strings */
	new_str(0);
}

