/*
 *      stack.c - A general purpose stack 
 *		Part of the usbsync utility 
 *
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "stack.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#ifdef MULTISTACK
/* Multistack include dependencies */
#  include "../file.h"
#  include "../error/usbs_errors.h"
#  include "../sync.h"
/* Names of stack types */
const char *stack_type_names[] = {"NORMAL_STACK", "FILE_STACK", "ERROR_STACK", "SYNCDATA_STACK"};
#endif


/* Stack malloc */
void *smalloc(size_t s) {
	void *mem;
	if( (mem = malloc(s)) == NULL )
		fprintf(stderr, "stack: memory allocation failed");	
	return mem;
} 

/**
 * Create and return new stack
 */
STACK* stack_new() {
	STACK *s = (STACK*)smalloc(sizeof(STACK));
	s->top = NULL;
	s->count = 0;
#ifdef MULTISTACK
	s->stack_type = NORMAL_STACK;
#endif
	return s;
}

/** 
 * Destroy and free stack
 */
void stack_destroy(STACK **s) {
	if( !*s ) {
		fprintf(stderr, "stack: free(): invalid stack\n");
		return;
	} else {
		/* traverse and remove stack elements */
		for( ; (*s)->count > 0; stack_pop(*s, 1) );
		free(*s);
		*s = NULL;
	}
}

/**
 * Returns 0 if successfully pushed onto stack, otherwise -1
 */
int stack_push(STACK *s, void *data) {
	STACK_ELEMENT *e;
	assert(s != NULL && data != NULL);
	if(s == NULL) {
		fprintf(stderr, "stack: invalid stack");
		return -1;
	}
	e = s->top;
	s->top = smalloc(sizeof(STACK_ELEMENT));
	if( s->top == NULL )
		return -1;
	s->top->next = e;
	s->top->data = data;
	s->count++;
	return 0;
}

/**
 * Pop top element, return 0 on success, -1 on failure.
 * If destroy is > 0 then data on the stack will be 
 * entirely free'd. If destory is not set, they will 
 * remain in memory.
 */
int stack_pop(STACK *s, int destroy) {
	STACK_ELEMENT *e;
	
	if(s == NULL) {
		fprintf(stderr, "\nstack: invalid stack");
		return -1;
	}
	
	/* Empty stack needs no popping */
	if( s->count < 1 || s->top == NULL) return -1;
	
	if( destroy ) {
		
#define TOPDEL  if( s->top->data != NULL ) { \
					free(s->top->data); \
					s->top->data = NULL; \
				}
		/* Different types of stacks need different kinds of cleanups, so 
		 * so decide here what needs to be done.
		 */				
#ifdef MULTISTACK
		switch(s->stack_type) {
			case NORMAL_STACK:
				TOPDEL
			break;
			case FILE_STACK:
				file_destroy((USBS_FILE**)&s->top->data);
			break;
			case ERROR_STACK:
				clean_error_message((ERROR_MESSAGE*)s->top->data);
			break;
			case SYNCDATA_STACK:
				syncdata_destroy((USBS_SYNCDATA**)&s->top->data);
			break;
		}
#else
			TOPDEL
#endif
		
	}
	
	e = s->top->next; 
	free(s->top);
	s->top = e;
	
	--s->count;
	
	return 0;
}

/**
 * If stack contains strings this function checks whether *str is
 * contained on stack. Return 1 if so, 0 if not, and -1 if an error was
 * encountered.
 */
int stack_str_pushed(STACK *s, const char *str) {
	STACK_ELEMENT *e;
	if(s == NULL) {
		fprintf(stderr, "\nstack: invalid stack");
		return -1;
	}
	for( e = s->top; e; e = e->next ) {
		if(!e->data) continue;
		if(!strncmp(str, e->data, strlen(str))) return 1;
	}
	return 0;
}
