/*
 *      status.c - Termcap percentage status display
 * 		Part of the usbsync utility 
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */


/* NOTE: This file does only get compiled if HAVE_TERMCAP
 * is not defined by configure (see settings.h and Makefile.am).
 */

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#  if STDC_HEADERS
#    include <string.h>
#  else
size_t strlen(const char *);
#  endif
#else
#  include <string.h>
#endif

#include "status.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <curses.h>
#include <term.h>

#define FLN __FILE__, __LINE__

/* Terminal lines and columns */
extern int cols;
extern int lins;

/* Used to avoid calling setupterm more than once */
static BOOL term_setup; 

/**
 * Returns a reference to a STATUS struct on success, otherwise NULL.
 */
STATUS* status_new(FILE* term_stream, int max, int start) {
	BOOL err = FALSE;
	STATUS *s = malloc(sizeof(STATUS));
	char *cuu1;
	if( s == NULL ) {
		fprintf(stderr, "\nstatus: (%s:%d): memory allocation failed\n", 
		FLN); 
		return NULL;
	} 
	if( max < 0 ) {
		fprintf(stderr, "\nstatus: (%s:%d): invalid max=%d value\n", 
		FLN, max); 
		err = TRUE;
	} else if ( start < 0 || start > max ) {
		fprintf(stderr, "\nstatus: (%s:%d): invalid start=%d value\n", 
		FLN, start);
		err = TRUE;
	}  else if ( term_stream == NULL ) {
		fprintf(stderr, "\nstatus: (%s:%d): corrput terminal stream (%#x)\n", 
		FLN, (unsigned int)term_stream);
		err = TRUE;
	}
	else {
		/* set struct */
		s->term_stream = term_stream, s->max = max, s->current = start,
		s->skip_fraction = 0, s->skip_count = 0;
		/* init term, parse terminfo*/
		if( !term_setup ) {
			setupterm((char*)0, fileno(term_stream), (int*)0); /* (int*)0 means exit on error) */
			term_setup = TRUE;	
		}
		if( (cuu1 = tigetstr("cuu1")) <= (char*)0 ) {
			fprintf(stderr, "\nstatus: (%s:%d): unable to retrieve termcap \"cuu1\"\n", 
			FLN);
			err = TRUE;
		}
		tputs(cuu1, 1, &putchar);
	} 
	if( err ) free(s), s = (STATUS *)NULL;
	return s;
}

/**
 * Destroys a status type
 */
void status_destroy(STATUS *s) {
	char *cud1;
	if(s == NULL) { 
		fprintf(stderr, "\nstatus: (%s:%d): cant't destroy status (status==NULL)\n", 
		FLN); 
		return;
	}
	if( (cud1 = tigetstr("cud1")) <= (char*)0 ) {
			fprintf(stderr, "\nstatus: (%s:%d): can't retrieve termcap \"cud1\"\n", 
			FLN);
			return;
	}
	tputs(cud1, 1, &putchar);
	free(s);
}


/**
 * Status output at the right hand side, that is, at the end of the 
 * current line.
 * Returns 0 on success, -1 on failure.
 * NOTE: status_new(...) must have been called before to init terminal.
 */
int status_print(STATUS *s, const char *prnt, ...) {
	char *hpa, *phpa, print[128];
	
	va_list l;	
	va_start(l, prnt);
	
	if( strlen(prnt) >= sizeof(print) ) {
		fprintf(stderr, "\nstatus: (%s:%d): print string too long\n", 
		FLN); 
		return -1;
	}
	
	/* check if term was set up */
	if( strnames == NULL ) {
		fprintf(stderr, "\nstatus: (%s:%d): printinf failed, terminfo was not initialized\n", 
		FLN); 
		return -1;
	}
	
	/* check STATUS structure */
	if(s == NULL) { 
		fprintf(stderr, "\nstatus: (%s:%d): printing failed (status==NULL)\n", 
		FLN); 
		return -1;
	}
	
	if( s->term_stream == NULL ) {
		fprintf(stderr, "\nstatus: (%s:%d): corrupt terminal stream (stream==NULL)\n", 
		FLN); 
		return -1;
	}
	
	if( (hpa  = tigetstr("hpa"))  <= (char*)0 ) {
		fprintf(stderr, "\nstatus: (%s:%d): could not retrieve termcap \"hpa\"\n", 
		FLN); 
		return -1;
	}
	
	/* assemble output string */
	vsprintf(print, prnt, l);
	
	if( (phpa = tparm(hpa, cols-strlen(print))) == NULL ) {
		fprintf(stderr, "\nstatus: (%s:%d): tparm error for \"hpa\"\n", 
		FLN); 
		return -1;
	}
	
	/* send formatted string to terminal*/
	tputs(phpa, 1, &putchar);
	
	/* output status percentage */
	fprintf(s->term_stream, "%s", print);
	fflush(s->term_stream);
	
	va_end(l);
	
	return 0;
	
}

/**
 * Attempts to update and output status percentage at the end of the
 * current line.
 */
int status_update(STATUS *s, int current) {
	char percp[16];
		
	if(s == NULL) { 
		fprintf(stderr, "\nstatus: (%s:%d): updating failed (status==NULL)\n", 
		FLN); 
		return -1;
	}
	
	s->current = current;
	
	/* only update when fractional part of max is reached (if set)*/
	if( s->skip_count < s->skip_fraction ) {
		++s->skip_count;	
		return 0;
	} else {
		s->skip_count = 0; /* reset skip count */
	}
	
	
	sprintf(percp, "[%d%%]", 
	(int)(((long double)s->current/(long double)s->max)*100.0));
	
	status_print(s, percp);
	
	return 0;
}

