/*
 *      sync.c - Routines for synchronizing files and directories
 * 		Part of the usbsync utility 
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <settings.h>
#endif

#include "sync.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h> 
#include <sys/types.h>
#include <errno.h>
#include <libgen.h>
#include <utime.h>

/* Local includes */
#include "common.h"
#include "options/options.h"
#include "stack/stack.h"
#include "debug/debug.h"
#include "file.h"
#include "error/usbs_errors.h"
#include "path.h"

#ifdef HAVE_TERMCAP
#  include "status/status.h"
#endif

/* Output strings for status */
#define _UPDATE_FAILED_STRING     "[failed]"
#define _UPDATE_SKIPPED_STRING    "[skipped]"
#define _UPDATE_SYMLINK_STRING    "[symlinked]"
#define _UPDATE_DELETED_STRING    "[removed]"
#define _UPDATE_DIRECTORY_UPDTD   "[updated]"
#define _UPDATE_DIRECTORY_CREAT   "[created]"
#define _UPDATE_DIRECTORY_DIR     "[dir]"

/* Output strings for times and misc*/
#define _UPTIME_NON_EXISTING  "non-existing"
#define TIME_FORMAT           "%D-%T"
#define _SYMLINK_MARKER       "(symlink)"
#define _DIR_MARKER           "(dir)"	  

/* Other functional macros */
/* Clear a line */
#define LINE_CLEAR() \
	move_cursor(UP, 0), move_cursor(CLR_CMPLT, 0)
/* Clear multiple line */
#define N_LINE_CLEAR(N) \
	while( N-- > 0 ) LINE_CLEAR();
/* Is passed a path, returns a symlink marking */
#define SYMLINK_MARKING(file)  ((is_symlink(file, 1)) ? (_SYMLINK_MARKER) : (EY))
/* Macro is passed a expand_type enum */
#define DIR_MARKING(X) ( (X == NORMAL_DIR || X == EMPTY_DIR || X == UPDATE_DIR) ?\
						  ( _DIR_MARKER" ") : (EY) )

/**
 * Initializes a typedefed USBS_SYNCDATA structure and returnes a reference
 * to it.
 */

USBS_SYNCDATA* syncdata_new() {
	USBS_SYNCDATA *s = usbs_malloc(sizeof(USBS_SYNCDATA));
	s->sync_files = s->prefix_ign = s->suffix_ign = NULL;
	s->host = s->usb = (char*)NULL;
	s->flags = 0;
	return s;
}

/**
 * Destroys a USBS_SYNCDATA object.
 */
void syncdata_destroy(USBS_SYNCDATA **s) {
	if( !*s )
		return;
	if( (*s)->usb ) free( (*s)->usb );
	if( (*s)->host ) free( (*s)->host );
	if( (*s)->sync_files ) stack_destroy(&(*s)->sync_files);
	if( (*s)->prefix_ign ) stack_destroy(&(*s)->prefix_ign);
	if( (*s)->suffix_ign ) stack_destroy(&(*s)->suffix_ign);
	free(*s);
	*s = NULL;
}


/**
 * Recursively ensures that directory passed as argument in initial
 * call will finally exist when this function exists. It makes all created dirs
 * be ownedr by user as specified in .usbsync file on usb.
 * Return values are as follows:
 *  0  - dir created
 *  1  - dir already exists
 *  2  - dir skipped (when `-p` opted)
 *  3  - dir only check (`-c` set)
 * -1  - an error occured
 */
static int subdir_create(enum expand_type direction, const char *dir, USER_INFO *uinfo) {
	char parent_dir[PATH_LENGTH], qst[256];
	
	verbose(VERBOSE_HIGH, "Checking directory:\n\"%s\"", PN_(dir, 0));
	
	if( !*dir ) {
		error_out(ERROR_ERROR_DIRREC, 0, FLN);
		return -1;
	}
	
	/* `-c` option set */
	if( _option_check(_OPTION_CHECK_ONLY) ) {
		return 3;
	}
		
	/* If dir exists we are fine */
	if( chk_existence(dir) ) { 
		verbose(VERBOSE_HIGH, "Directory does already exist:\n\"%s\"", dir);
		return 1;
	}
	
	/* If parent directory does exists create this directory 
	 * NOTE: Obsolete in the sense described in the sentence above 
	 *       as directories are pushed after 
	 *       their contents, meaning that parent dirs will already 
	 * 		 exist anyways. 
	 */
	get_directory(parent_dir, dir);
	if( access(parent_dir, 0) != 0 ) {
		verbose(VERBOSE_HIGH, "Parent directory does not exist:\n"
		"parent_dir=\"%s\"\ndir=\"%s\"", 
		PN_(parent_dir, 0), PN_(dir, 0));
		return subdir_create(direction, parent_dir, uinfo);	
	}
	
	dir_prompt:
	
	/* `-p` option set */
	if( _option_check(_OPTION_UPDATE_PROMPT) ||
		(_option_check(_OPTION_REMOV_SNGL) && !chk_existence(dir)) ) {
		sprintf(qst, "Create directory: \"%s\" ?", basename((char*)dir));
		switch( prompt_question(qst, "YNIS", &msg) ) {
			case 'N':
				verbose(VERBOSE_MID, "Directory not handled:\n\"%s\"\v", PN_(dir, 0));
				return 2;
			break;
			case 'S':
				sprintf(qst, "Really skip dir \"%s\" plus the rest?", basename((char*)dir));
				if( prompt_question(qst, "YN", &msg) == 'Y' ) {
					LINE_CLEAR();
					_option_set(_OPTION_CHECK_ONLY);
					return 3;
				}
				goto dir_prompt;
			break;
			/* Further information about directory location */
			case 'I':
				msg("Location: \"%s\"", PN_(dir, 0));
				getchar();
				LINE_CLEAR();
				goto dir_prompt;
			break;
			case (char)0:
				error_out(ERROR_ERROR_USRCHO, 0, FLN);
				return -1;
			break;	
		}
	}
	
	/* Now create directory */
	if( mkdir(dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0 ) {
		error_out(ERROR_ERROR_MKDIR, errno, FLN, dir);
		return -1;
	} 
	
	verbose(VERBOSE_LOW, "Directory created:\n\"%s\"\v", dir);
	
	return 0;
}

		
/** 
 * Returns 0 if both exist.
 * Returns 1 if only source exists. 
 * Returns -1 if source does not exist or if an error occured.
 * Writes formatted time string to buffers oldtime and newtime. 
 */
static int frmt_time_info(const char *src, const char *dst, char *oldtime, char *newtime) {

	struct stat st;
	struct tm *tim;
	
	*oldtime = *newtime = 0;
	
	/* Source file information */
	if( !chk_existence(src) ) return -1;
	
	if( usbs_stat((char*)src, &st) < 0 ) return -1;

	tim = gmtime(&st.st_mtime);
	
	if( strftime(newtime, 128, TIME_FORMAT, tim) == 0 ) {
		error_out(ERROR_ERROR_STRFRMT, 0, FLN, newtime);
		return -1;
	}
	
	/* Destination file information */
	if( !chk_existence(dst) ) {
		strcpy(oldtime, _UPTIME_NON_EXISTING);
		return 1;
	}
	
	if( usbs_stat((char*)dst, &st) < 0 ) return -1;
	tim = gmtime(&st.st_mtime);
	
	if( strftime(oldtime, 128, TIME_FORMAT, tim) == 0 ) {
		error_out(ERROR_ERROR_STRFRMT, 0, FLN, oldtime);
		return -1;
	}
	
	return 0;
}
		
/*
 * This enum determines how "much too old" checkings are to be handled when
 * updating a file or directory in update_file() below. "Much too old"
 * means the program should prompt for confirmation if a destination file
 * is much much older (TOO_OLD_TIME) than its source file.
 */
#define TOO_OLD_TIME 180*DAY_SECONDS /* Roughly half a year */
static enum { ALWAYS_CHECK, ALWAYS_SKIP, ALWAYS_CONTINUE } too_old_handling;

/**
 * Updates files and directories given by source and destination.
 * Return 0 on success, -1 on failure.
 */
static int update_file(
	enum usbs_filetype ftype,
	enum expand_type direction,
	const char *src, 
	const char *dest, 
	const char *src_name, 
	const char *dest_name, 
	time_t src_mtime,
	time_t dst_mtime,
	uid_t uid, gid_t gid) 
{

/* Used when skipping this file somewhere in this routine */
#define SKIP_FILE  move_cursor(UP, 0);\
				   wrp_status_print(_UPDATE_SKIPPED_STRING);\
				   goto success
/* The opposite direction */
#define OPPOSITE(DIRECT) (DIRECT == HD_TO_USB ? USB_TO_HD : HD_TO_USB)
/* Clear line and move cursor up */
#define CLEAN_AND_UP move_cursor(CLR_CMPLT, 0), move_cursor(UP, 1)
	
	char backupfile[PATH_LENGTH], linkaddr[PATH_LENGTH];
	char file_size[128], prompt[512], prompt_options[512],
		 oldtime[PATH_LENGTH], newtime[128];
	int tmp;
#ifdef HAVE_TERMCAP
	STATUS s = { .term_stream = stdout };
#endif
	USER_INFO uinfo = { .uid = uid, .gid = gid };
	
	*newtime = *oldtime = 0;
	
	/* FIXME: Not sure if this is necessary; but sometimes weird 
	 * things are happening if they are not zero'd out 
	 */
	bzero(oldtime, sizeof oldtime);
	bzero(newtime, sizeof newtime);
	
	/* Critical code section begins:
	   This means when catching any fatal signal they are caught by the signal
	   handler after data has been copied or whatever. That is to ensure 
	   data safety.
	 */	
	__start_critical();
	
	/* If directory only needs updating (dest already exists); skip: 
	   This is done in second round when ftype == UPDATE_DIR.
	 */
	if( (ftype == NORMAL_DIR || ftype == EMPTY_DIR) && chk_existence(dest) )
		goto success;
		
	/* If dest is non-existent and dest is a directory, 
	   we are in the second round, and `-p` or `-g` is set we skip here 
	 */
	if( !chk_existence(dest) && ftype == UPDATE_DIR && 
		(_option_check(_OPTION_UPDATE_PROMPT) || _option_check(_OPTION_REMOV_SNGL)) )
		goto success;
	
	/* Output synchonization info */
	__out();
	msg("%s:\"%s\"%s => %s:\"%s\"%s\v", 
	src_name, FN_(basename((char*)src)), SYMLINK_MARKING(src),
	dest_name, FN_(basename((char*)dest)), SYMLINK_MARKING(dest));
	
	/* Check if we have write access to destination directory */
	get_directory(oldtime, dest);
	if( access(oldtime, F_OK | W_OK | X_OK) < 0 && !_option_check(_OPTION_CHECK_ONLY) ) {
		error_out(ERROR_ERROR_ACCESS, 0, FLN, PN_(oldtime, 0));
		move_cursor(UP, 2);
#ifdef HAVE_TERMCAP
		status_print(&s, _UPDATE_FAILED_STRING);
#endif
		goto failure;
	}
	
	/* Warn if target file or directory are much much older than source */
	if( dst_mtime > 0 && src_mtime - dst_mtime > TOO_OLD_TIME ) {
		if( too_old_handling == ALWAYS_SKIP ) {
			SKIP_FILE;
		}
		else if( too_old_handling == ALWAYS_CHECK ) {
			warning_cnt++;
			msg("WARNING: Source is much more recent than destination\v");
			if( frmt_time_info(src, dest, oldtime, newtime) < 0 )
				error_out(ERROR_ERROR_FRMTERR, 0, FLN); 
			else {
				msg("%s => %s\v", oldtime, newtime);				
			}
			
			/* Skip all means skip all that would cause this warning */
			switch( prompt_question("Continue / Skip / Continue All / Skip All ?", "CSAL", &msg) ) {
				/* Skip */
				case 'S':
					CLEAN_AND_UP;
					SKIP_FILE;
				break;
				/* Skip all */
				case 'L':
					too_old_handling = ALWAYS_SKIP;
					CLEAN_AND_UP;
					SKIP_FILE;
				break;
				/* Continue all */
				case 'A':
					too_old_handling = ALWAYS_CONTINUE;
				break;
				/* Continue */
				case 'C':
				default:
					/* FIXME: Default fault */;
				break;
			}
			CLEAN_AND_UP;
		}
		/* Remaining case is ALWAYS_CONTINUE; this means just continue from here */
	}
	 
	/* Handle DIRECTORIES first now */
	if( ftype == NORMAL_DIR || ftype == EMPTY_DIR || ftype == UPDATE_DIR ) {		
		dir_again:
		/* Try to create/handle directory; decide according to what happened */
		switch( subdir_create(direction, dest, &uinfo) ) {
			/* An error occurred */
			case -1:
				move_cursor(UP, 0);
				wrp_status_print(_UPDATE_FAILED_STRING);
				goto failure;
			break;
			/* Created */
			case 0: 
				move_cursor(UP, 1);
				/* "Created" status */
				wrp_status_print(_UPDATE_DIRECTORY_CREAT);
				goto success;
			break;
			/* Exists */
			case 1:
				/* Prompt if dir should be updated if `-p` or `-g` is set */
				if( _option_check(_OPTION_UPDATE_PROMPT)
					/* FIXME: figure out if dir was a single before and prompt here */
					/* || (_option_check(_OPTION_REMOV_SNGL) && chk_existence(dest)) */
					) {
					switch( prompt_question("Would you like to update directory?", "YNS", &msg) ) {
						case 'N':
							SKIP_FILE;
						case 'S':
							if( prompt_question("Really skip this directory plus the rest?", 
												"YN", &msg) == 'Y' ) {
								_option_set(_OPTION_CHECK_ONLY);					
							}
							goto dir_again;
						break;
					}
						
				}
				/*
				FIXME: Redundant
				If option `-o` is set, do not attempt to update
				if( _option_check(_OPTION_IGN_OWNER) ) {
					move_cursor(UP, 1);
					wrp_status_print(_UPDATE_DIRECTORY_DIR);
					goto success;
				}
				*/
				
				/* Update directory modifcation time, owner, and mode */
				if( (equate_am_time(src, dest, 0) | /* Set modification time */
				    usbs_chown(direction, dest, uinfo.uid, uinfo.gid) | /* Set owner */
				    mode_equate(direction, dest, src)) == 0 /* Set (copy) permission bits */ ) {
				    	
				    /* Sucessfully handled dir */
				    move_cursor(UP, 1);
					wrp_status_print(_UPDATE_DIRECTORY_UPDTD);
					++dirs_synced;
					goto success;
					
				} else {
					
					/* An error occured */
					move_cursor(UP, 1);
					wrp_status_print(_UPDATE_FAILED_STRING);
					goto failure;
					
				}
			break;
			/* Skip directory */
			case 2:
			/* Check only */
			case 3:
			
				++dirs_skipped;
				move_cursor(UP, 0);
				wrp_status_print(_UPDATE_SKIPPED_STRING);
				
				if( (ftype == NORMAL_DIR || ftype == EMPTY_DIR) && 
					!chk_existence(dest) ) {
					msg("Directory creation necessary\v");
				} else {
					if( frmt_time_info(src, dest, oldtime, newtime) < 0 ) {
						error_out(ERROR_ERROR_FRMTERR, 0, FLN); 
					} else {
						msg("Direct. update necces.: %s => %s\v", oldtime, newtime);
					}
				}
				
				goto success;
				
			break;
		}
		
	} else
	
	/* Handle FILES now */
	
	if( _option_check(_OPTION_UPDATE_PROMPT) 	|| /* Prompt before action if -p option set */
		_option_check(_OPTION_CHECK_ONLY)		|| /* If -c */
		(!chk_existence(dest) && _option_check(_OPTION_REMOV_SNGL)) /* If singles are to be checked */ 
	) {
		
		/* Information string */
		tmp = frmt_time_info(src, dest, oldtime, newtime);
		
		/* Only src exists */
		if( tmp == 1 ) {
			strcpy(prompt_options, "YNISD");
		/* Both exist */
		} else if ( tmp == 0 ){
			strcpy(prompt_options, "YNISR");
		/* Error */
		} else {
			error_out(ERROR_ERROR_FRMTERR, 0, FLN);
		}
		
		/* If source is dangling link prompting doesnt make sense */
		if( is_dangling_symlink(src) && !_option_check(_OPTION_CPY_SYMLINKS) ) {
			msg("Source is a dangling link. Try the `-l' option.\v\v");
			move_cursor(UP, 1);
			SKIP_FILE;
			goto success;
		}
		
		/* If check only (-c) we jump to check here */
		if( _option_check(_OPTION_CHECK_ONLY) ) goto check_only;
			
		prompt:
		sprintf(prompt, "Replace: %s => %s ?", newtime, oldtime);
		
		/* Pose question */
		switch( prompt_question(prompt, prompt_options, &msg) ) {
			case 'N':
				/* Skip if "N" pressed */
				++files_skipped;
				SKIP_FILE;
			break;
			/* Replace newer file with old file */
			case 'R':
				msg("Swapped source and destination\v");
				return update_file(ftype, OPPOSITE(direction), dest, src, dest_name, src_name, 
								   dst_mtime, src_mtime, uid, gid);
			break;
			/* Remove old file if there is only a signle copy. */
			case 'D':
				sprintf(prompt, "Really delete: \"%s\" ?", FN_(basename((char*)src)));
				if( prompt_question(prompt, "YN", &msg) == 'Y' ) {
					if( unlink(src) < 0 ) {
						error_out(ERROR_ERROR_UNLINK, errno, FLN, PN_(src, 1));
						goto failure;
					} else {
						msg("Removed from %s: \"%s\"", src_name, FN_(basename((char*)src)));
						move_cursor(UP, 1);
						wrp_status_print(_UPDATE_DELETED_STRING);
						++files_removed;
						goto success;
					}
				} else goto prompt;
			break;
			/* Further file information: show were file is located; show link targets */
			case 'I':
				
				tmp = 2;
				
				/* Source info */
				bytes_to_human(file_size, true_filesize(src));
				get_directory(backupfile, src);
				msg("Source: Size: %s, Dir.: \"%s\"\v", file_size, PN_(backupfile, 0));
				
				/* Symlink info */
				if( is_symlink(src, 1) ) {
					symlink_info(linkaddr, "Source link target", src); 
					msg("%s\v", PN_(linkaddr, 0) );
					tmp++;
				}
				
				/* Destination info */
				if(true_filesize(dest) > 0) bytes_to_human(file_size, true_filesize(dest));
				else strcpy(file_size, "none");
				get_directory(backupfile, dest);
				msg("Desti.: Size: %s, Dir.: \"%s\"\v", file_size, PN_(backupfile, 0));	
				
				/* Symlink info */
				if( is_symlink(dest, 1) ) {
					symlink_info(linkaddr, "Desti. link target", dest);
					msg("%s\v", PN_(linkaddr, 0));
					tmp++;
				}

				/* Clear line; back to prompt */
				usbs_getchar();				
				N_LINE_CLEAR(tmp);
				goto prompt;
			break;
			/* Skip everything else */
			case 'S':
				if( prompt_question("Really skip this step as well as everything else?", "YN", 
				&msg) == 'Y' ) {
					move_cursor(UP, 1);
					move_cursor(CLR_CMPLT, 0);
					_option_set(_OPTION_CHECK_ONLY);
					goto check_only;
				} else goto prompt;
			break;
			case (char)0:
				error_out(ERROR_ERROR_USRCHO, 0, FLN);
				goto failure;
			break;
		}
	}
	
	/* If dest is dangling symlink w/o existent destination file skip possible backup */
	if( !is_dangling_symlink(dest) ) {
		
		/* If `-s` has been set, create backup file of dest (old file) before overwriting */
		if( _option_check(_OPTION_SAVE_SYNC) && access(dest, F_OK) > -1 ) {
			verbose(VERBOSE_HIGH, "Attempting to create backup");
			get_directory(backupfile, dest);
			sprintf(backupfile, "%s/.%s"BACKUPFILE_SUFFIX, backupfile, basename((char *)dest));
			verbose(VERBOSE_LOW, "Creating backup ... \nold=\"%s\"\nback=\"%s\"", 
			PN_(dest, 1), PN_(backupfile, 1));
			if( cpy_file(dest, backupfile, uid, gid, FALSE) < 0 ) {
				error_out(ERROR_ERROR_BACKUP, 0, FLN, PN_(dest, 1), PN_(backupfile, 1));
			}
		}
		
	} else verbose(VERBOSE_MID, "Dangling link detected, possible backups skipped:\n\"%s\"", 
	PN_(dest, 1));
		 	
	
	/* If -l is set, try to copy symlinks instead of files */
	if( _option_check(_OPTION_CPY_SYMLINKS) && is_symlink(src, TRUE) ) { 
		tmp = cpy_symlink(src, dest, uid, gid, TRUE);
		/* An error occurred. */
		if( tmp < 0 ) {
			usbs_error(FALSE, 0, FLN, 
			"Copying symlink failed\nSource:\"%s\"\nDestination:\"%s\"", PN_(src, 1), PN_(dest, 1));
			move_cursor(UP, 0);
			wrp_status_print(_UPDATE_FAILED_STRING);
			goto failure;
		}
		/* Symlink got synced. */
		else if( tmp == 0 ) {
			move_cursor(UP, 1);
			wrp_status_print(_UPDATE_SYMLINK_STRING);
		}
	/* Check if source is a dangling symlink */
	} else if( is_dangling_symlink(src) ) { 
		verbose(VERBOSE_MID, "Source is a dangling symlink, skip copying:\n"
		"src=\"%s\"\ndst=\"%s\"", PN_(src, 1), PN_(dest, 1));
		msg("Source is a dangling link. Use `-l' to copy anyways.\v");
		++files_skipped;
		SKIP_FILE;
	}
	/* In any other case source might have been a symlink, but there is data to be copied */
	else {
		/* Now copy file */
		verbose(VERBOSE_LOW, "Copying ...\nsrc=\"%s\"\ndest=\"%s\"", 
		PN_(src, 1), PN_(dest, 1));
		if( cpy_file(src, dest, uid, gid, FALSE) < 0 ) {
			usbs_error(FALSE, 0, FLN, 
			"Copying failed:\nSrc: \"%s\"\nDest:\"%s\"", PN_(src, 1), PN_(dest, 1));
			move_cursor(UP, 0);
			wrp_status_print(_UPDATE_FAILED_STRING);
			goto failure;
		}
		
		/* Set right file `user and group` ownership */
		if( usbs_chown(!strcmp(dest_name, MASS_STORAGE_NAME) ? HD_TO_USB : USB_TO_HD, 
			dest, uid, gid) < 0 ) {
			error_store(ESTACK_ERRORS, ERROR_ERROR_USBCHOWN, errno, FLN,
			(char*)dest, uid, gid);
		}
	}
	
	++files_synced;	
	
	success:
	__end_critical();
	__out();
	return 0;	
	
	failure:
	__end_critical();
	__out();
	return -1;	
	
	/* `-c` option has been set (this here is for "real" files only; no dirs) */
	check_only:
	++files_skipped;
	move_cursor(UP, 1);
	wrp_status_print(_UPDATE_SKIPPED_STRING);
	if( tmp >= 0 )
	msg("Update necessary: %s => %s\v", newtime, oldtime);
	goto success;
}

/**
 * Checks whether PREFIX/POSTFIX on STACK fixes match with fname.
 * The compare function (cmpfunc) passed is of a strncmp-like form.
 * This could either be .. for .. and .
 */
static BOOL psfix_check(
	int (*cmpfunc)(const char*, const char*, size_t), 
	STACK *fixes, 
	const char *fname) 
{	
	STACK_ELEMENT *e;
	char *ign;
	
	if( !fixes ) return FALSE;
	
	for( e = fixes->top; e; e = e->next ) {
		ign = (char*)e->data;
		/* Check if match according to cmpfunc */
		if( cmpfunc(ign, fname, strlen(ign)) == 0 ) 
			return TRUE;
	}
	
	return FALSE;	
}

/**
 * Is handed a stack of file syncdata structures and attempts to sync files on the
 * stack syncdata->sync_files with each other. Prefices and suffices on the
 * stacks prefix_ign as well as suffix_ign are ignored. 
 * See parser.c for more detail.
 */
void sync_files(STACK *syncdata_stack) {
	
#define SKIPPING  msg("Synchronization entry skipped\v");\
				  continue
	
	BOOL file_ignored, destroy, dirs_syn, synced, skipping;
	USER_INFO uinfo;
	USBS_SYNCDATA *sync_data;
	USBS_FILE *f = NULL;
	STACK *file_stack, *prefix_ignore, *suffix_ignore;
	STACK *dir_stack = stack_new();
	
	dirs_syn = synced = skipping = FALSE;
	
	dir_stack->stack_type = FILE_STACK;

	__block();
	msg("Synchronizing files and directories %s... ", 
	_option_check(_OPTION_SAVE_SYNC) ? "(Save Synchronization) " : ""); 
	
	for( ; syncdata_stack->count > 0; stack_pop(syncdata_stack, 1) ) {
		
		/* Take next syncdata structure from stack */
		sync_data = (USBS_SYNCDATA *)syncdata_stack->top->data;
		
		/* Assign file stack */
		file_stack = sync_data->sync_files;
		
		/* Assign prefix ignore stack */
		prefix_ignore = sync_data->prefix_ign;
		
		/* Assign suffix ignore stack */
		suffix_ignore = sync_data->suffix_ign;
		
		__block();
		putchar('\n');
		msg("\"%s\" <=> \"%s\":\v", 
			PNN_(sync_data->usb,  RELPATH, 0.75 * MAX_PATHN_LEN, 0),
			PNN_(sync_data->host, ABSPATH, 0.75 * MAX_PATHN_LEN, 0));
			
		synced = dirs_syn = FALSE;
		
		sync_prompt:
		if( _option_check(_OPTION_UPDATE_PROMPT) && !skipping ) {
			switch( prompt_question("Would you like to skip this entry?", "YNS", &msg) ) {
				case 'Y':
					SKIPPING;
				break;
				case 'S':
					if( prompt_question("Really skip all synchronizaton entries?", "YN", &msg) == 'N' ) {
						move_cursor(CLR_CMPLT, 0);
						goto sync_prompt;
					}
					skipping = TRUE;
				break;
			}
		}
		
		if( skipping ) {
			SKIPPING;
		}
		
		/* If tape_archive; switch to tar routines and continue */
		/// ... TODO
		
		synchronize:		
		if( !file_stack ) {
			continue;
		}
		
		/* Take each single file from the stack */
		for( ; file_stack->count > 0; stack_pop(file_stack, destroy) ) {
			f = ((USBS_FILE*)file_stack->top->data);
			
			/* When popping, destory file by default. File does not get popped
			 * if file is a dir; then it goes to dir_stack.
			 */
			destroy = TRUE;
			
			/* Time measuring, always per line */
			time_needed += __stop_timer();
			__start_timer();
			
			uinfo.uid = f->uid, uinfo.gid = f->gid;
			
			/* Prefix/suffix ignoring doesn't apply do directories */
			if( !is_directory(f) ) {
			
				/* Check if we should ignore file because 
				 * its suffix or prefix are disabled.
				 * For this purpose the two comparison functions strncmp and
				 * strrncmp are utilized in psfix_check (see for further detail).
				 */
				file_ignored = FALSE;
				
				/* Check if usb file name is in IGNORED prefices */
				if( psfix_check(&strncmp, prefix_ignore, basename(f->usb_fname)) )
					++file_ignored;
				/* Check if usb file name is in IGNORED suffices */
				else if( psfix_check(&strrncmp, suffix_ignore, basename(f->usb_fname)) )
					++file_ignored;
				/* Check if hd (host) file name is in IGNORED prefices */
				else if( psfix_check(&strncmp, prefix_ignore, basename(f->host_fname)) )
					++file_ignored;
				/* Check if hd (host) file name is in IGNORED suffices */
				else if( psfix_check(&strrncmp, suffix_ignore, basename(f->host_fname)) )
					++file_ignored;
					
				if( file_ignored ) {
					verbose(VERBOSE_LOW, "File ignored due to its "
					"prefix/suffix:\nusb=\"%s\"\nhd=\"%s\"", 
					PN_(f->usb_fname, 1), PN_(f->host_fname, 1));
					continue;
				}	
			}
			
			/* If file is a dir push file onto separat dir stack.
			 * This is needed to update directories in the end,
			 * because changes in dirs cause its mtime and atime to
			 * be altered.
			 */ 
			
			/* Delete older file if existing and replace with more recent one*/
			if( f->usb_time > f->host_time ) {
			
				/* Dump usbs file on error */
				if( (update_file(f->file_type, USB_TO_HD, f->usb_fname,
					f->host_fname, MASS_STORAGE_NAME, 
					this_hostname, f->usb_time, f->host_time, f->uid, f->gid) < 0) &&
					 verbosity_check(VERBOSE_MID) ) 
					 
					dump_usbs_file(f);
					
				synced = TRUE;
					
			} else if( f->host_time > f->usb_time ) {
				
				/* Dump usbs file on error */
				if( (update_file(f->file_type, HD_TO_USB, f->host_fname, f->usb_fname, this_hostname, 
					 MASS_STORAGE_NAME, f->host_time, f->usb_time, f->uid, f->gid) < 0) &&
					 verbosity_check(VERBOSE_MID) ) 
					 
					dump_usbs_file(f);
				
				synced = TRUE; 
					
			} else {
				verbose(VERBOSE_LOW, "Already up-to-date (mtime=%d)\nusb:\"%s\"\n%s:\"%s\"", 
				f->usb_time, PN_(f->usb_fname, 1), this_hostname, PN_(f->host_fname, 1));
			}
			
			/* File was a directory; push onto directory stack */
			if( is_directory(f) && !dirs_syn) {
				verbose(VERBOSE_MID, "Pushing directory onto directory stack:\nusb:\"%s\"\nhdd:\"%s\"",
				PN_(f->usb_fname, 0), PN_(f->host_fname, 0));
				
				f->file_type = UPDATE_DIR;
				
				stack_push(dir_stack, f);
				destroy = FALSE;
			}
			
		}
		
		/* Initiate second round of syncing, this time with directory stack.
		 * This is important so that all directories are proplery set
		 * up to date after the actual synchronization of files has finished.
		 */
		if( !dirs_syn ) {
			file_stack = dir_stack;
			++dirs_syn;
			goto synchronize;
		}
		
		if( !synced ) msg("Everything in synchronization entry already up-to-date");
	
	}
	
	/* Cleanups */
	stack_destroy(&dir_stack);
}

