/*
 *      usbsync.c - A storage volume file and directory synchronization utility
 *      
 *      Copyright 2008 Sören Wellhöfer <soeren.wellhoefer@gmx.net>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

/* Autoconf compile-time settings */
#ifdef HAVE_CONFIG_H
#  include <settings.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>

#include "common.h"
#include "error/usbs_errors.h"
#include "options/options.h"
#include "debug/debug.h"
#include "parser.h"
#include "stack/stack.h"
#include "sync.h"
#include "args/args.h"
#include "signals/signals.h"

/* Give package information if no config header */
#ifndef HAVE_CONFIG_H
	#define PACKAGE  			"usbsync"
	#define VERSION  			"testbuild"
	#define PACKAGE_BUGREPORT   "maintainer"
#endif

/* Program identifier */
const char *progname = PACKAGE;
const char *version  = VERSION;
const char *descr    = "Mass storage file synchronization utility";
const char *license  = "GPL";
const char *author   = "Soeren Wellhoefer (soeren.wellhoefer@gmx.net)";
const char *year     = "2008";
const char *info 	 = "GNU %s %s - %s\n\n"
				       "Copyright (C) %s Free Software Foundation, Inc.\n"
    			       "This is free software; see the source for copying conditions.  There is NO\n"
    				   "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
    				   "Written by %s\n\n\n";
const char *bugs     = "Please help by reporting bugs to <" PACKAGE_BUGREPORT ">.\n";

/* Usage string */
static const char *usage_str = 
	"Usage: %s [OPTIONS]... MOUNTPOINT\n"
	"%s - (%s v. %s under %s)\n\n"
	"Options:\n"
	"   -f, --usbsync-file    FILE     Specify to use FILE instead of the default\n" 
	"                                   file \".usbsync\" on the storage volume's\n" 
	"                                   mountpoint to read synchronization\n"
	"                                   information from.\n"
	"   -p, --prompt                  Interactive mode; a prompt will be displayed\n"
	"                                   before updating. This prompt does also\n"
	"                                   offer the possibility of replacing a more\n" 
	"                                   recent version of a file with an older one,\n"
	"                                   or to delete a single file without counterpart.\n"
	"   -g, --single                   When synchronizing, stop at each file or di-\n"
	"                                   rectory without counterpart and prompt what\n"
	"                                   action is to be taken.\n"
	"   -c, --check-only               No synchronization tasks will be performed;\n"		    
	"                                   only print which files and directories\n"
	"                                   need updating.\n"  
	"   -s, --save-sync                For each updated file there will a backup\n"
	"                                   copy according to the naming convention\n"
	"                                    \".<filename>.ubs\" be present in the same\n"
	"                                    directory.\n"     
	"   -d, --dir-recursion   DEPTH    When synchronizing directories, specifies at\n"
	"                                   which recusive sub-directory level (depth)\n"
	"                                   the program should stop. 0 means only files\n"
	"                                   in the top-level directory.\n"
	"   -l, --symlinks                 Copy symbolic links themselves rather than\n"
	"                                  the files or directories they point to; do\n"
	"                                   not follow or dereference.\n"
	"   -o, --ignore-owner             Enable this option if you neither want to\n"
	"                                   have any particual user and group\n" 
	"                                   ownership nor any access permissions set\n"
	"                                   for the objects synchronized onto the\n"
	"                                   mass volume. This could come in handy\n"
	"                                   in connection with certain file system\n"
	"                                   that do not support the Unix concept\n"
	"                                   of file or directory ownership (like vfat).\n"
	"   -e, --no-error                 Suppress all error and warning messages.\n"
	"   -v, --verbose         LEVEL    Verbose output, whereby verbosity level:\n"
	"                                   0 < LEVEL < 4\n"
	"   -q, --quiet                    Do not output anything.\n"
	"   -h, --help                     Print this message.\n"
	"   --version                      Print program information\n\n";

/* Opting variables */
sig_atomic_t verbosity; /* Verbosity from 0 (no verbose output) to 3 */
BOOL quiet;    			/* Do not output anything */

/* Globals */
char usbsyncfile_name[PATH_LENGTH]; /* Default is USBSYNCFILE=".usbsync" */
char this_hostname[HOST_NAME_MAX];    /* Hostname of current machine */
char this_username[LOGIN_NAME_MAX];	/* Username of current executing user */
int  dir_recursion;         /* Recursive depth when syncing directories. Initialized with -1. */

/**
 * Prints program information
 */
void prog_info() {
	printf(info, progname, version, descr, year, author);
	puts(bugs);
}    				  

/** 
 * Print a newline character 
 */
static void newline() {
	putc('\n', stdout);
}

/**
 * Outputs the usage string.
 * If error is TRUE then something with the usage string given
 * was wrong and the program exits here.
 */
static void usage(BOOL error) {
       fprintf(error ? stderr : stdout, usage_str, 
       progname, descr, progname, version, license);
       puts(bugs);
       exit( error ? EXIT_FAILURE : EXIT_SUCCESS );
}

/**
 * Everytime the program exits this function is called
 * to really make sure that everything that was somehow allocated
 * through an initilizer function will be free'd eventually by calling
 * the approriate deinitializer.
 */
static void cleanups() {
	__deinit_path_display();
	__deinit_errors();
}

/**
 * usbsync program entry point.
 */
int main(int argc, char **argv) {
	
	
	/*printf("%s\n", 
	PNN_("/home/wave/wfawf/wfafwaf/awfwafawf/awfwafa/afwaf/qwwqqfwqfwfwfewfweew", ABSPATH, 20, 1));
	
	return 0;*/

	char short_opts[128], option;
	STACK *syncdata;
	sigset_t segv_mask;

	/* Long options */
	struct option long_opts[] = {
              {"usbsync-file",     required_argument, NULL, 'f'},
              {"save-sync",        no_argument,       NULL, 's'},
              {"prompt",           no_argument,       NULL, 'p'}, 
              {"dir-recursion",    required_argument, NULL, 'd'}, 
              {"check-only",       no_argument,       NULL, 'c'},
              {"symlinks",         no_argument,       NULL, 'l'},
              {"ignore-owner",     no_argument,       NULL, 'o'},
              {"no-error",         no_argument,       NULL, 'e'},
              {"single",           no_argument,       NULL, 'g'},
              {"verbose",          required_argument, NULL, 'v'},
              {"quiet",            no_argument,       NULL, 'q'},
              {"help",             no_argument,       NULL, 'h'},
              {"version",          no_argument,       NULL,  1},
              {0, 0, 0, 0}
    };
    
    /* Initialize application to have two error stacks: normal errors and warnings 
     * FIXME: Error queues would be much more logical. Should follow. 
     */
    __init_errors(2);
    
    /* Set segmentation faults to be caught */
    sigfillset(&segv_mask);
	SET_SIGNAL(SIGSEGV, &segv_mask, 0);
	
	/* Enable signal catching; caught signals will have their entry point here */
	SIGNAL_CATCHING();
	
	/* Handle sigsegv when caught */
	CATCH_SIGNAL(SIGSEGV) {
		error_out(ERROR_FATAL_SEGV, 0, FLN, GET_SIGINFO()->si_addr);
	}
    
    /* Cleanups on exist */
    atexit(&cleanups);
    
    /* Make shortopts from longopts array */
    make_shortops(short_opts, long_opts); 
    
    /* Default settings */
    verbosity = quiet = FALSE;
    dir_recursion = -1;
	strcpy(usbsyncfile_name, USBSYNCFILE);
	
	/* Initialize options set by arguments to "no option set yet" */
	_init_options();
	
    /* Parse argument list */
    while( (option = getopt_long(argc, argv, short_opts, long_opts, NULL)) > -1 ) {
            switch(option) {
				case 'h':
					usage(FALSE);
				break;   
				/* Unknown arg */           
				case '?':	
					usage(TRUE);
				break;
				case 'f':
					if( strlen(optarg) < NAME_LENGTH )
						strcpy(usbsyncfile_name, optarg);
					else
						error_out(ERROR_FATAL_USBSYNC, 0, FLN, optarg);
				break;
				case 'q':
					quiet = TRUE;
					verbosity = FALSE;
				break;
				case 'v':
					if( !quiet ) {
						verbosity = atoi(optarg);
						if( verbosity < 1 || verbosity > 3 ) {
							error_out(ERROR_ERROR_VLEVEL, 0, 0, 0, verbosity);
							usage(TRUE);
						}
					}
				break;
				/* --prompt */
				case 'p':
					_option_set(_OPTION_UPDATE_PROMPT);
				break;
				/* --check-only */
				case 'c':
					_option_set(_OPTION_CHECK_ONLY);
				break;
				/* --save-sync */
				case 's':
					_option_set(_OPTION_SAVE_SYNC);
				break;
				/* --dir-recursion */
				case 'd':
					dir_recursion = atoi(optarg);
					if( dir_recursion < 0 ) {
						error_out(ERROR_ERROR_DREC, 0, 0, 0, dir_recursion);
						usage(TRUE);
					}
				break;
				/* --symlinks */
				case 'l':
					_option_set(_OPTION_CPY_SYMLINKS);
				break;
				/* --no-error*/
				case 'e':
					_option_set(_OPTION_SUPPR_ERR);
				break;
				/* --ignore-owner */
				case 'o':
					_option_set(_OPTION_IGN_OWNER);
				break;
				/* --single-file */
				case 'g':
					_option_set(_OPTION_REMOV_SNGL);
				break;
				/* --version */
				case  1:
					prog_info();
					return EXIT_SUCCESS;
				break;
			}
	}
	
    /* At least one argument, that is, the mass medium mount point, must exist */
    if( optind == argc ) {
    	error_out(ERROR_ERROR_MEDIUM, 0, 0, 0);
		usage(TRUE);
	}
	
	/* Invalid option use */
	if( (_option_check(_OPTION_UPDATE_PROMPT) || _option_check(_OPTION_REMOV_SNGL)) && 
		 quiet ) {
		quiet = FALSE;
		error_out(ERROR_ERROR_TGETHER, 0, 0, 0);
		usage(1);
	}
	
	if(verbosity_check(VERBOSE_HIGH)) _option_print();
	
   	/* Initialize block spacing, statistics count, 
   	 * terminfo db use, pathname shortening, signal handlers 
   	 * and register print_conclusion to be called when exitting.
   	 */
	__init();
	__init_statistics();
	__init_termcap();
	__init_path_display();
	__init_sighandlers();
	atexit(&print_conclusion);
	
	if( quiet ) atexit(&newline);
	
	/* Get hostname and username */
	hostname(this_hostname);
	username(this_username);
	
	/* Retrieve stack of USBS_SYNCDATA struct to be synchronized */
	syncdata = parse_usbsync( *(argv + optind) );
	
	/* Dump usbsync data stack on high verbosity */
	if( verbosity_check(VERBOSE_HIGH) ) dump_stack(syncdata, 0);
	
	if( !syncdata->count ) msg("No scan results\v");

	/* 
	 * Attempt to synchronize stack of USBS_SYNCDATA
	 * If this fails the program will be abnormally exited within the function.
	 */ 
	sync_files(syncdata);
	
	/* Clean up stack of syncdata */
	stack_destroy(&syncdata);
	
	/* Output errors/warnings on stacks */
	errors_out(ESTACK_ERRORS);
	errors_out(ESTACK_WARNINGS);
	
	return EXIT_SUCCESS;
}
